#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-
#   
#   PYTHON MODULE:     MKI18N.PY
#                      =========
# 
#   Abstract:         Make Internationalization (i18n) files for an application.
#   Copyright Pierre Rouleau. 2003. Released to public domain.
#   Last update: Saturday, November 8, 2003. @ 15:55:18.
#   File: ROUP2003N01::C:/dev/python/mki18n.py
# 
#   Update history:
#   - File updated: Thursday, June 12, 2007 by Cody Precord
#   - File created: Saturday, June 7, 2003. by Pierre Rouleau
#   - 07/12/07: Make it work with current wx code and clean up code CJP
#   - 10/06/03 rcs : RCS Revision 1.1  2003/06/10 10:06:12  PRouleau
#   - 10/06/03 rcs : RCS Initial revision
#   - 23/08/03 rcs : RCS Revision 1.2  2003/06/10 10:54:27  PRouleau
#   - 23/08/03 P.R.: [code:fix] : The strings encoded in this file are encode 
#                                 in iso-8859-1 format.  Added the encoding
#                    notification to Python to comply with Python's 2.3 PEP 263.
#   - 23/08/03 P.R.: [feature:new] : Added the '-e' switch which is used to 
#                                    force the creation of the empty English 
#                                    .mo file.
#   - 22/10/03 P.R.: [code] : incorporated utility functions in here 
#                             to make script self sufficient.
#   - 05/11/03 rcs : RCS Revision 1.4  2003/10/22 06:39:31  PRouleau
#   - 05/11/03 P.R.: [code:fix] : included the unixpath() in this file.
#   - 08/11/03 rcs : RCS Revision 1.5  2003/11/05 19:40:04  PRouleau
# 
#   RCS $Log: $
# 
# 
# -----------------------------------------------------------------------------
"""                                
mki18n allows you to internationalize your software.  You can use it to 
create the GNU .po files (Portable Object) and the compiled .mo files
(Machine Object).

mki18n module can be used from the command line or from within a script (see 
the Usage at the end of this page).

    Table of Contents
    -----------------
    makePO()  -- Build the Portable Object file for the application --
    catPO()   -- Concatenate one or several PO files with the application 
                 domain files.
    makeMO()   -- Compile the Portable Object files into the Machine Object 
                  stored in the right location.
    printUsage   -- Displays how to use this script from the command line --
    Scriptexecution      -- Runs when invoked from the command line --

NOTE:  this module uses GNU gettext utilities.

You can get the gettext tools from the following sites:

   - `GNU FTP site for gettetx`_ where several versions 
     (0.10.40, 0.11.2, 0.11.5 and 0.12.1) are available.
     Note  that you need to use `GNU libiconv`_ to use this. Get it from the 
     `GNU
     libiconv  ftp site`_ and get version 1.9.1 or later. Get the Windows .ZIP
     files and install the packages inside c:/gnu. All binaries will be stored
     inside  c:/gnu/bin.  Just  put c:/gnu/bin inside your PATH. You will need
     the following files: 

      - `gettext-runtime-0.12.1.bin.woe32.zip`_ 
      - `gettext-tools-0.12.1.bin.woe32.zip`_
      - `libiconv-1.9.1.bin.woe32.zip`_ 


.. _GNU libiconv:   http://www.gnu.org/software/libiconv/
.. _GNU libiconv ftp site:  http://www.ibiblio.org/pub/gnu/libiconv/
.. _gettext-runtime-0.12.1.bin.woe32.zip:
        ftp://ftp.gnu.org/gnu/gettext/gettext-runtime-0.12.1.bin.woe32.zip 
.. _gettext-tools-0.12.1.bin.woe32.zip: 
..      ftp://ftp.gnu.org/gnu/gettext/gettext-tools-0.12.1.bin.woe32.zip 
.. _libiconv-1.9.1.bin.woe32.zip: 
        http://www.ibiblio.org/pub/gnu/libiconv/libiconv-1.9.1.bin.woe32.zip

"""
# -----------------------------------------------------------------------------
# Module Import
# -------------
# 
import os
import sys
import wx
# -----------------------------------------------------------------------------
# Global variables
# ----------------
#

__author__ = "Pierre Rouleau"
__version__ = "$Revision: 70029 $"

# -----------------------------------------------------------------------------

def getlanguageDict():
    """Get a dictionary of the available languages from wx"""
    lang_dict = {}
    app = wx.App()
    for lang in [x for x in dir(wx) if x.startswith("LANGUAGE")]:
        i = wx.Locale(wx.LANGUAGE_DEFAULT).GetLanguageInfo(getattr(wx, lang))
        if i:
            lang_dict[i.CanonicalName] = i.Description
    return lang_dict

# -----------------------------------------------------------------------------
# m a k e P O ( )  -- Build the Portable Object file for the application --
# ^^^^^^^^^^^^^^^
#
def makePO(app_dir,  app_domain=None, verbose=0) :
    """Build the Portable Object Template file for the application.

    makePO builds the .pot file for the application stored inside 
    a specified directory by running xgettext for all application source 
    files. It finds the name of all files by looking for a file called 
    'app.fil'. 
    If this file does not exists, makePo raises an IOError exception.
    By default the application domain (the application
    name) is the same as the directory name but it can be overridden by the
    'applicationDomain' argument.

    makePO always creates a new file called messages.pot.  If it finds files 
    of the form app_xx.po where 'app' is the application name and 'xx' is one 
    of the ISO 639 two-letter language codes, makePO resynchronizes those 
    files with the latest extracted strings (now contained in messages.pot). 
    This process updates all line location number in the language-specific
    .po files and may also create new entries for translation (or comment out 
    some).  The .po file is not changed, instead a new file is created with 
    the .new extension appended to the name of the .po file.

    By default the function does not display what it is doing.  Set the 
    verbose argument to 1 to force it to print its commands.
    """

    if app_domain is None:
        app_name = fileBaseOf(app_dir, withPath=0)
    else:
        app_name = app_domain
    curr_dir = os.getcwd()
    os.chdir(app_dir)                    
    if not os.path.exists('app.fil'):
        raise IOError(2,'No module file: app.fil')

    # Steps:                                  
    #  Use xgettext to parse all application modules
    #  The following switches are used:
    #  
    #   -s : sort output by string content 
    #        (easier to use when we need to merge several .po files)
    #   --files-from=app.fil : The list of files is taken from the file: app.fil
    #   --output= : specifies the name of the output file 
    #               (using a .pot extension)
    cmd = 'xgettext -s --from-code=utf-8 --no-wrap --files-from=app.fil --output=Editra.pot'
    if verbose: 
        print cmd
    os.system(cmd)                                                

    lang_dict = getlanguageDict()
    for lang_code in lang_dict.keys():
        if lang_code == 'en':
            pass
        else:
            lang_po_filename = "%s_%s.po" % (app_name, lang_code)
            if os.path.exists(lang_po_filename):
                cmd = 'msgmerge -s --no-wrap "%s" Editra.pot > "%s.new"' % \
                      (lang_po_filename, lang_po_filename)
                if verbose: 
                    print cmd
                os.system(cmd)
    os.chdir(curr_dir)

# -----------------------------------------------------------------------------
# c a t P O ( ) 
# -- Concatenate one or several PO files with the application domain files. --
# ^^^^^^^^^^^^^
#
def catPO(app_dir, listOf_extraPo, app_domain=None, \
          target_dir=None, verbose=0) :
    """Concatenate one or several PO files with the application domain files.
    """

    if app_domain is None:
        app_name = fileBaseOf(app_dir, withPath=0)
    else:
        app_name = app_domain
    curr_dir = os.getcwd()
    os.chdir(app_dir)

    lang_dict = getlanguageDict()

    for lang_code in lang_dict.keys():
        if lang_code == 'en':
            pass
        else:
            lang_po_fname = "%s_%s.po" % (app_name, lang_code)
            if os.path.exists(lang_po_fname):
                fileList = ''
                for fname in listOf_extraPo:
                    fileList += ("%s_%s.po " % (fname, lang_code))
                cmd = "msgcat -s --no-wrap %s %s > %s.cat" % (lang_po_fname, \
                                                              fileList, \
                                                              lang_po_fname)
                if verbose: 
                    print cmd
                os.system(cmd)
                if target_dir is None:
                    pass
                else:
                    mo_targetDir = "%s/%s/LC_MESSAGES" % (target_dir, lang_code)
                    cmd = "msgfmt --output-file=%s/%s.mo %s_%s.po.cat" % \
                           (mo_targetDir, app_name, app_name, lang_code)
                    if verbose: 
                        print cmd
                    os.system(cmd)
    os.chdir(curr_dir)

# -----------------------------------------------------------------------------
# m a k e M O ( ) Compile the POfiles into the MO stored in the right location.
# ^^^^^^^^^^^^^^^
# 
def makeMO(applicationDirectoryPath, targetDir='./locale', 
           applicationDomain=None, verbose=0, forceEnglish=0) :
    """Compile the Portable Object files into the Machine Object stored in the 
    right location.

    makeMO converts all translated language-specific PO files located inside 
    the  application directory into the binary .MO files stored inside the 
    LC_MESSAGES sub-directory for the found locale files.

    makeMO searches for all files that have a name of the form 'app_xx.po' 
    inside the application directory specified by the first argument.  The 
    'app' is the application domain name (that can be specified by the 
    applicationDomain argument or is taken from the directory name). The 'xx' 
    corresponds to one of the ISO 639 two-letter language codes.

    makeMo stores the resulting files inside a sub-directory of `targetDir`
    called xx/LC_MESSAGES where 'xx' corresponds to the 2-letter language
    code.
    """
    if targetDir is None:
        targetDir = './locale'
    if verbose:
        print "Target directory for .mo files is: %s" % targetDir

    if applicationDomain is None:
        applicationName = fileBaseOf(applicationDirectoryPath, withPath=0)
    else:
        applicationName = applicationDomain
    currentDir = os.getcwd()
    os.chdir(applicationDirectoryPath)

    languageDict = getlanguageDict()

    for langCode in languageDict.keys():
        if (langCode == 'en') and (forceEnglish==0):
            pass
        else:
            langPOfileName = "%s_%s.po" % (applicationName, langCode)
            if os.path.exists(langPOfileName):
                mo_targetDir = "%s/%s/LC_MESSAGES" % (targetDir, langCode) 
                if not os.path.exists(mo_targetDir):
                    mkdir(mo_targetDir)
                cmd = 'msgfmt --output-file="%s/%s.mo" "%s_%s.po"' % \
                      (mo_targetDir, applicationName, applicationName, langCode)
                if verbose: 
                    print cmd
                os.system(cmd)
    os.chdir(currentDir)
   
# -----------------------------------------------------------------------------
# printUsage -- Displays how to use this script from the command line --
# ^^^^^^^^^^^^^^^^^^^
#
def printUsage(errorMsg=None):
    """Displays how to use this script from the command line."""
    print """
    ###########################################################################
    #   mki18n : Make internationalization files.                             #
    #            Uses the GNU gettext system to create PO (Portable Object)   #
    #            files from source code, compile PO into MO (Machine Object)  #
    #            files.                                                       #
    #            Supports C, C++, and Python source files.                    #
    #                                                                         #
    #   Usage: mki18n {OPTION} [appDirPath]                                   #
    #                                                                         #
    #   Options:                                                              #
    #     -e               : When -m is used, forces English .mo file         #
    #                        creation                                         #
    #     -h               : prints this help                                 #
    #     -m               : make MO from existing PO files                   #
    #     -p               : make PO, update PO files: Creates a new          #
    #                        messages.pot file. Creates a dom_xx.po.new for   #
    #                        every existing language specific .po file.       #
    #                        ('xx' stands for the ISO639 two-letter language  #
    #                        code and 'dom' stands for the application domain #
    #                        name).  mki18n requires that you write a         #
    #                       'app.fil' file  which contains the list of all    #
    #                        source code to parse.                            #
    #     -v               : verbose (prints comments while running)          #
    #     --domain=appName : specifies the application domain name.           #
    #                        By default the directory name is used.           #
    #     --moTarget=dir : specifies the directory where .mo files are stored.#
    #                      If not specified, the target is './locale'         #
    #                                                                         #
    #   You must specify one of the -p or -m option to perform the work. You  #
    #   can specify the path of the target application.  If you leave it out  #
    #   mki18n will use the current directory as the application main         #
    #   directory.                                                            #
    ###########################################################################
    """
    if errorMsg:
        print "\n   ERROR: %s" % errorMsg

# -----------------------------------------------------------------------------
# f i l e B a s e O f ( )         -- Return base name of filename --
# ^^^^^^^^^^^^^^^^^^^^^^^
# 
def fileBaseOf(filename, withPath=0) :
    """fileBaseOf(filename,withPath) ---> string

    Return base name of filename. The returned string never includes the 
    extension.
    Use os.path.basename() to return the basename with the extension.  The 
    second argument is optional.  If specified and if set to 'true' (non zero)
    the string returned contains the full path of the file name.  Otherwise the
    path is excluded.

    [Example]
    >>> fn = 'd:/dev/telepath/tvapp/code/test.html'
    >>> fileBaseOf(fn)
    'test'
    >>> fileBaseOf(fn)
    'test'
    >>> fileBaseOf(fn,1)
    'd:/dev/telepath/tvapp/code/test'
    >>> fileBaseOf(fn,0)
    'test'
    >>> fn = 'abcdef'
    >>> fileBaseOf(fn)
    'abcdef'
    >>> fileBaseOf(fn,1)
    'abcdef'
    >>> fn = "abcdef."
    >>> fileBaseOf(fn)
    'abcdef'
    >>> fileBaseOf(fn,1)
    'abcdef'
    """            
    pos = filename.rfind('.')             
    if pos > 0:
        filename = filename[:pos]
    if withPath:
        return filename
    else:
        return os.path.basename(filename)
# -----------------------------------------------------------------------------
# m k d i r ( )   -- Create a directory (and possibly the entire tree) --
# ^^^^^^^^^^^^^
# 
def mkdir(directory) :
    """Create a directory (and possibly the entire tree).

    The os.mkdir() will fail to create a directory if one of the
    directory in the specified path does not exist.  mkdir()
    solves this problem.  It creates every intermediate directory
    required to create the final path. Under Unix, the function 
    only supports forward slash separator, but under Windows and MacOS
    the function supports the forward slash and the OS separator (backslash
    under windows).
    """ 
    # translate the path separators
    directory = unixpath(directory)

    # build a list of all directory elements
    elem_list = filter(lambda x: len(x) > 0, directory.split('/'))
    the_len = len(elem_list)

    # if the first element is a Windows-style disk drive
    # concatenate it with the first directory
    if elem_list[0].endswith(':'):
        if the_len > 1:
            elem_list[1] = elem_list[0] + '/' + elem_list[1]
            del elem_list[0]      
            the_len -= 1     

    # if the original directory starts at root,
    # make sure the first element of the list 
    # starts at root too
    if directory[0] == '/':     
        elem_list[0] = '/' + elem_list[0]

    # Now iterate through the list, check if the 
    # directory exists and if not create it
    the_dir = ''
    for i in range(the_len):
        the_dir += elem_list[i]
        if not os.path.exists(the_dir):
            os.mkdir(the_dir)
        the_dir += '/'   
      
# -----------------------------------------------------------------------------
# u n i x p a t h ( ) -- Return a path name that contains Unix separator. --
# ^^^^^^^^^^^^^^^^^^^
# 
def unixpath(path) :
    r"""Return a path name that contains Unix separator.

    [Example]
    >>> unixpath(r"d:\test")
    'd:/test'
    >>> unixpath("d:/test/file.txt")
    'd:/test/file.txt'
    >>> 
    """
    path = os.path.normpath(path)
    if os.sep == '/':
        return path
    else:
        return path.replace(os.sep, '/')

# ----------------------------------------------------------------------------- 
# S c r i p t   e x e c u t i o n  -- Runs when invoked from the command line --
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# 
if __name__ == "__main__":
    import getopt     # command line parsing
    argc = len(sys.argv)
    if argc == 1:
        printUsage('Missing argument: specify at least one of -m or -p '
                   '(or both).')
        sys.exit(1)
    # If there is some arguments, parse the command line
    valid_opts = "ehmpv"
    valid_lopts = ['domain=', 'moTarget=']             
    option = {}
    option['forceEnglish'] = 0
    option['mo'] = 0
    option['po'] = 0        
    option['verbose'] = 0
    option['domain'] = None
    option['moTarget'] = None
    try:
        optionList, pargs = getopt.getopt(sys.argv[1:], valid_opts, valid_lopts)
    except getopt.GetoptError, e:
        printUsage(e[0])
        sys.exit(1)       
    for (opt, val) in optionList:
        if  (opt == '-h'):    
            printUsage()
            sys.exit(0) 
        elif (opt == '-e'):
            option['forceEnglish'] = 1
        elif (opt == '-m'):
            option['mo'] = 1
        elif (opt == '-p'):
            option['po'] = 1
        elif (opt == '-v'):
            option['verbose'] = 1
        elif (opt == '--domain'):
            option['domain'] = val
        elif (opt == '--moTarget'):
            option['moTarget'] = val
    if len(pargs) == 0:
        app_dir_path = os.getcwd()
        if option['verbose']:
            print "No project directory given. Using current one:  %s" % \
                                                                    app_dir_path
    elif len(pargs) == 1:
        app_dir_path = pargs[0]
    else:
        printUsage(('Too many arguments (%u). Use double quotes if you have '
                   'space in directory name') % len(pargs))
        sys.exit(1)
    if option['domain'] is None:
        # If no domain specified, use the name of the target directory
        option['domain'] = fileBaseOf(app_dir_path)
    if option['verbose']:
        print "Application domain used is: '%s'" % option['domain']
    if option['po']:
        try:
            makePO(app_dir_path, option['domain'], option['verbose'])
        except IOError, e:
            printUsage(e[1] + ('\n   You must write a file app.fil that '
                              'containsthe list of all files to parse.'))
    if option['mo']:
        makeMO(app_dir_path, option['moTarget'], option['domain'], 
               option['verbose'], option['forceEnglish'])
    sys.exit(1)            
            
# -----------------------------------------------------------------------------



###############################################################################
# Name: progress.py                                                           #
# Purpose: Define Progress 4gl syntax for highlighting and other features     #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
Syntax highlighting definition for Progress 4GL programming language.

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _progress.py 68798 2011-08-20 17:17:05Z CJP $"
__revision__ = "$Revision: 68798 $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc

# Local Imports
import synglob
import syndata
import _sql

#-----------------------------------------------------------------------------#

# Reserved Progress 4GL keywords
PROG_KW = (0, "accumulate active-window add alias all alter ambiguous analyze " 
              "and any apply as ascending assign at attr-space authorization "
              "auto-return available background before-hide begins bell "
              "between blank break btos by call can-do can-find centered "
              "character check chr clear clipboard col colon color column "
              "column-label columns compiler connected control count-of "
              "cpstream create ctos current current-changed current-language "
              "current-window current_date cursor database dataservers "
              "dbcodepage dbcollation dbname dbrestrictions dbtaskid dbtype "
              "dbversion dde deblank debug-list debugger decimal decimals "
              "declare def default default-noxlate default-window define "
              "delete delimiter descending dictionary disable disconnect disp "
              "display distinct dos down drop editing enable encode entry "
              "error-status escape etime except exclusive exclusive-lock "
              "exclusive-web-user exists export false fetch field fields "
              "file-information fill find find-case-sensitive find-global "
              "find-next-occurrence find-prev-occurrence find-select "
              "find-wrap-around first first-of focus font form format frame "
              "frame-col frame-db frame-down frame-field frame-file "
              "frame-index frame-line frame-name frame-row frame-value from "
              "from-chars from-pixels gateways get-byte get-codepages "
              "get-collations get-key-value getbyte global go-on go-pending "
              "grant graphic-edge group having header help hide import in "
              "index indicator input input-output insert integer into is "
              "is-attr-space join kblabel key-code key-function key-label "
              "keycode keyfunction keylabel keys keyword label last last-event "
              "last-key last-of lastkey ldbname leave library like "
              "line-counter listing locked lookup machine-class map member "
              "message message-lines mouse mpe new next next-prompt no "
              "no-attr-space no-error no-fill no-help no-hide no-labels "
              "no-lock no-map no-message no-pause no-prefetch no-undo "
              "no-validate no-wait not null num-aliases num-dbs num-entries "
              "of off old on open opsys option or os-append os-command "
              "os-copy os-create-dir os-delete os-dir os-drives os-error "
              "os-rename os2 os400 output overlay page page-bottom page-number "
              "page-top parameter pause pdbname persistent pixels preprocess "
              "privileges proc-handle proc-status process program-name "
              "Progress prompt prompt-for promsgs propath proversion put "
              "put-byte put-key-value putbyte query query-tuning quit r-index "
              "rcode-information readkey recid record-length rectangle release "
              "reposition retain retry return return-value revert revoke run "
              "save schema screen screen-io screen-lines scroll sdbname search "
              "seek select self session set setuserid share-lock shared "
              "show-stats skip some space status stream stream-io string-xref "
              "system-dialog table term terminal text text-cursor "
              "text-seg-growth this-procedure time title to today top-only "
              "trans transaction trigger triggers trim true underline undo "
              "unformatted union unique unix up update use-index use-revvideo "
              "use-underline user userid using v6frame value values variable "
              "view view-as vms wait-for web-context window window-maximized "
              "window-minimized window-normal with work-table workfile write "
              "xcode xref yes _cbit _control _list _memory _msg _pcontrol "
              "_serial-num _trace "
              "repeat transaction for each end finf where if then else skip "
              "close"
)

# Progress 4GL Types
PROG_TYPES = (1, "char character int integer format var variable log logical "
                 "da date")

# Progress 4GL Operators
PROG_OP = (7, "absolute accelerator across add-first add-last advise alert-box "
              "allow-replication ansi-only anywhere append appl-alert-boxes "
              "application as-cursor ask-overwrite attachment auto-end-key "
              "auto-endkey auto-go auto-indent auto-resize auto-zap "
              "available-formats average avg backwards base-key batch-mode "
              "bgcolor binary bind-where block-iteration-display border-bottom "
              "border-bottom-chars border-bottom-pixels border-left "
              "border-left-chars border-left-pixels border-right "
              "border-right-chars border-right-pixels border-top "
              "border-top-chars border-top-pixels both bottom box "
              "box-selectable browse browse-header buffer buffer-chars "
              "buffer-lines button buttons byte cache cache-size can-query "
              "can-set cancel-break cancel-button caps careful-paint "
              "case-sensitive cdecl character character_length charset checked "
              "choose clear-selection close code codepage codepage-convert "
              "col-of colon-aligned color-table column-bgcolor column-dcolor "
              "column-fgcolor column-font column-label-bgcolor "
              "column-label-dcolor column-label-fgcolor column-label-font "
              "column-of column-pfcolor column-scrolling combo-box command "
              "compile complete connect constrained contents context "
              "context-popup control-container control-form convert-to-offset "
              "convert count cpcase cpcoll cpinternal cplog cpprint cprcodein "
              "cprcodeout cpterm crc-value create-control "
              "create-result-list-entry create-test-file current-column "
              "current-environment current-iteration current-result-row "
              "current-row-modified current-value cursor-char cursor-line "
              "cursor-offset data-entry-return data-type date date-format day "
              "db-references dcolor dde-error dde-id dde-item dde-name "
              "dde-topic debug decimal default-button default-extension "
              "defer-lob-fetch define defined delete-char delete-current-row "
              "delete-line delete-selected-row delete-selected-rows "
              "deselect-focused-row deselect-rows deselect-selected-row "
              "design-mode dialog-box dialog-help dir disabled display-message "
              "display-type double drag-enabled drop-down drop-down-list dump "
              "dynamic echo edge edge-chars edge-pixels editor empty end-key "
              "endkey entered eq error error-column error-row event-type "
              "events exclusive-id execute exp expand extended extent external "
              "extract fetch-selected-row fgcolor file file-name file-offset "
              "file-type filename fill-in filled filters first-child "
              "first-column first-procedure first-tab-item fixed-only float "
              "focused-row font-based-layout font-table force-file foreground "
              "forult-row current-row-modified current-value cursor-char "
              "cursor-line cursor-offset data-entry-return data-type date "
              "date-format day db-references full-width full-width-chars "
              "full-width-pixels ge get get-blue-value get-char-property "
              "get-double get-dynamic get-file get-float get-green-value "
              "get-iteration get-license get-long get-message get-number "
              "get-pointer-value get-red-value get-repositioned-row "
              "get-selected-widget get-short get-signature get-size get-string "
              "get-tab-item get-text-height get-text-height-chars "
              "get-text-height-pixels get-text-width get-text-width-chars "
              "get-text-width-pixels get-unsigned-short grayed "
              "grid-factor-horizontal grid-factor-vertical grid-set grid-snap "
              "grid-unit-height grid-unit-height-chars grid-unit-height-pixels "
              "grid-unit-width grid-unit-width-chars grid-unit-width-pixels "
              "grid-visible gt handle height height-chars height-pixels "
              "help-context helpfile-name hidden hint horizontal hwnd image "
              "image-down image-insensitive image-size image-size-chars "
              "image-size-pixels image-up immediate-display index-hint "
              "indexed-reposition information init initial initial-dir "
              "initial-filter initiate inner inner-chars inner-lines "
              "insert-backtab insert-file insert-row insert-string insert-tab "
              "integer internal-entries is-lead-byte is-row-selected "
              "is-selected item items-per-row join-by-sqldb keep-frame-z-order "
              "keep-messages keep-tab-order key keyword-all label-bgcolor "
              "label-dcolor label-fgcolor label-font label-pfcolor labels "
              "languages large large-to-small last-child last-tab-item "
              "last-procedure lc le leading left left-aligned left-trim length "
              "line list-events list-items list-query-attrs list-set-attrs "
              "list-widgets load load-control load-icon load-image "
              "load-image-down load-image-insensitive load-image-up "
              "load-mouse-pointer load-small-icon log logical lookahead lower "
              "lt manual-highlight margin-extra margin-height "
              "margin-height-chars margin-height-pixels margin-width "
              "margin-width-chars margin-width-pixels matches max max-chars "
              "max-data-guess max-height max-height-chars max-height-pixels "
              "max-rows max-size max-value max-width max-width-chars "
              "max-width-pixels maximize maximum memory menu menu-bar "
              "menu-item menu-key menu-mouse menubar message-area "
              "message-area-font message-line min min-height min-height-chars "
              "min-height-pixels min-size min-value min-width min-width-chars "
              "min-width-pixels minimum mod modified modulo month "
              "mouse-pointer movable move-after-tab-item move-before-tab-item "
              "move-column move-to-bottom move-to-eof move-to-top multiple "
              "multiple-key multitasking-interval must-exist name native ne "
              "new-row next-column next-sibling next-tab-item next-value "
              "no-apply no-assign no-bind-where no-box no-column-scrolling "
              "no-convert no-current-value no-debug no-drag no-echo "
              "no-index-hint no-join-by-sqldb no-lookahead no-row-markers "
              "no-scrolling no-separate-connection no-separators no-underline "
              "no-word-wrap none num-buttons num-columns num-copies "
              "num-formats num-items num-iterations num-lines "
              "num-locked-columns num-messages num-results num-selected "
              "num-selected-rows num-selected-widgets num-tabs num-to-retain "
              "numeric numeric-format octet_length ok ok-cancel "
              "on-frame-border ordered-join ordinal orientation os-getenv "
              "outer outer-join override owner page-size page-width paged "
              "parent partial-key pascal pathname pfcolor pinnable "
              "pixels-per-column pixels-per-row popup-menu popup-only position "
              "precision preselect prev prev-column prev-sibling prev-tab-item "
              "primary printer-control-handle printer-setup private-data "
              "profiler Progress-source publish put-double put-float put-long "
              "put-short put-string put-unsigned-short query-off-end question "
              "radio-buttons radio-set random raw raw-transfer read-file "
              "read-only real recursive refresh refreshable replace "
              "replace-selection-text replication-create replication-delete "
              "replication-write request resizable resize retry-cancel "
              "return-inserted return-to-start-dir reverse-from right "
              "right-aligned right-trim round row row-markers row-of rowid "
              "rule rule-row rule-y save-as save-file screen-value scroll-bars "
              "scroll-delta scroll-horiz-value scroll-offset "
              "scroll-to-current-row scroll-to-item scroll-to-selected-row "
              "scroll-vert-value scrollable scrollbar-horizontal "
              "scrollbar-vertical scrolled-row-position scrolling "
              "se-check-pools se-enable-off se-enable-on se-num-pools "
              "se-use-message section select-focused-row select-next-row "
              "select-prev-row select-repositioned-row select-row selectable "
              "selected selected-items selection-end selection-list "
              "selection-start selection-text send sensitive "
              "separate-connection separators set-blue-value set-break "
              "set-cell-focus set-contents set-dynamic set-green-value "
              "set-leakpoint set-pointer-value set-property set-red-value "
              "set-repositioned-row set-selection set-size set-wait-state "
              "side-lab side-labe side-label side-label-handle side-labels "
              "silent simple single size size-chars size-pixels slider "
              "smallint sort source source-procedure sql sqrt start "
              "status-area status-area-font status-bar stdcall stenciled stop "
              "stopped stored-procedure string sub-average sub-count "
              "sub-maximum sub-menu sub-menu-help sub-minimum sub-total "
              "subscribe substitute substring subtype sum super "
              "suppress-warnings system-alert-boxes system-help tab-position "
              "tabbable target target-procedure temp-directory temp-table "
              "terminate text-selected three-d through thru tic-marks "
              "time-source title-bgcolor title-dcolor title-fgcolor title-font "
              "to-rowid toggle-box tool-bar top topic total trailing truncate "
              "type unbuffered unique-id unload unsubscribe upper use "
              "use-dict-exps use-filename use-text v6display valid-event "
              "valid-handle validate validate-condition validate-message "
              "variable vertical virtual-height virtual-height-chars "
              "use-filename use-text v6display valid-event valid-handle "
              "validate validate-condition validate-message variable vertical "
              "virtual-height virtual-height-chars widget-pool width "
              "width-chars width-pixels window-name window-state window-system "
              "word-wrap x-of y-of year yes-no yes-no-cancel _dcm")


#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for Progress 4GL""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_SQL)

    def GetKeywords(self):
        """Progress 4GL keyword specifications """
        return [PROG_KW, PROG_TYPES, PROG_OP]

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return _sql.SYNTAX_ITEMS

    def GetProperties(self):
        """Language properties for folding ect... """
        return [_sql.FOLD,]

    def GetCommentPattern(self):
        """Comment pattern """
        return [u'/*', u'*/']



###############################################################################
# Name: ed_i18n.py                                                            #
# Purpose: I18n utilities and services                                        #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
This file is a module for managing translations and the internationalization of
the program.

METHODS:
  - L{GetAvailLocales}: Returns a list of canonical names of available locales
  - L{GetLocaleDict}: Returns a dictionary consisting of canonical names for
                      keys and language ids for values.

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: ed_i18n.py 71189 2012-04-12 00:37:04Z CJP $"
__revision__ = "$Revision: 71189 $"

#--------------------------------------------------------------------------#
# Imports
import os
import wx
import wx.lib.langlistctrl as langlist
import wx.combo
import glob

# Editra Imports
import ed_glob

#----------------------------------------------------------------------------#
# Global Variables
OPT_NO_OP    = 0
OPT_DESCRIPT = 1

#----------------------------------------------------------------------------#

#---- Helper Functions used by the classes in this module ----#
def GetAvailLocales():
    """Gets a list of the available locales that have been installed
    for the editor. Returning a list of strings that represent the
    canonical names of each language.
    @return: list of all available local/languages available

    """
    avail_loc = list()
    loc = glob.glob(os.path.join(ed_glob.CONFIG['LANG_DIR'], "*"))
    for path in loc:
        the_path = os.path.join(path, "LC_MESSAGES", ed_glob.PROG_NAME + ".mo")
        if os.path.exists(the_path):
            avail_loc.append(os.path.basename(path))
    return avail_loc

def GetLocaleDict(loc_list, opt=OPT_NO_OP):
    """Takes a list of cannonical locale names and by default returns a
    dictionary of available language values using the canonical name as
    the key. Supplying the Option OPT_DESCRIPT will return a dictionary
    of language id's with languages description as the key.
    @param loc_list: list of locals
    @keyword opt: option for configuring return data
    @return: dict of locales mapped to wx.LANGUAGE_*** values

    """
    lang_dict = dict()
    for lang in [x for x in dir(wx) if x.startswith("LANGUAGE_")]:
        langId = getattr(wx, lang)
        langOk = False
        try:
            langOk = wx.Locale.IsAvailable(langId)
        except wx.PyAssertionError:
            continue

        if langOk:
            loc_i = wx.Locale.GetLanguageInfo(langId)
            if loc_i:
                if loc_i.CanonicalName in loc_list:
                    if opt == OPT_DESCRIPT:
                        lang_dict[loc_i.Description] = langId
                    else:
                        lang_dict[loc_i.CanonicalName] = langId
    return lang_dict

def GetLangId(lang_n):
    """Gets the ID of a language from the description string. If the
    language cannot be found the function simply returns the default language
    @param lang_n: Canonical name of a language
    @return: wx.LANGUAGE_*** id of language

    """
    if lang_n == "Default":
        # No language set, default to English
        return wx.LANGUAGE_ENGLISH_US
    lang_desc = GetLocaleDict(GetAvailLocales(), OPT_DESCRIPT)
    return lang_desc.get(lang_n, wx.LANGUAGE_DEFAULT)

#---- Language List Combo Box----#
class LangListCombo(wx.combo.BitmapComboBox):
    """Combines a langlist and a BitmapComboBox"""
    def __init__(self, parent, id_, default=None):
        """Creates a combobox with a list of all translations for the
        editor as well as displaying the countries flag next to the item
        in the list.

        @param default: The default item to show in the combo box

        """
        lang_ids = GetLocaleDict(GetAvailLocales()).values()
        lang_items = langlist.CreateLanguagesResourceLists(langlist.LC_ONLY, \
                                                           lang_ids)
        wx.combo.BitmapComboBox.__init__(self, parent, id_,
                                         size=wx.Size(250, 26),
                                         style=wx.CB_READONLY)
        for lang_d in lang_items[1]:
            bit_m = lang_items[0].GetBitmap(lang_items[1].index(lang_d))
            self.Append(lang_d, bit_m)

        if default:
            self.SetValue(default)

#-----------------------------------------------------------------------------#
if __name__ == '__main__':
    APP = wx.PySimpleApp(False)
    # Print a list of Canonical names useful for seeing what codes to
    # use when naming po files
    OUT = list()
    for LANG in [x for x in dir(wx) if x.startswith("LANGUAGE")]:
        LOC_I = wx.Locale.GetLanguageInfo(getattr(wx, LANG))
        if LOC_I:
            OUT.append((LOC_I.Description, LOC_I.CanonicalName))

    for LANG in sorted(OUT):
        print LANG



#-----------------------------------------------------------------------------
# Name:        stcspellcheck.py
# Purpose:     Spell checking for the wx.StyledTextControl using pyenchant
#
# Author:      Rob McMullen
#
# Created:     2008
# RCS-ID:      $Id: stcspellcheck.py 69791 2011-11-19 22:02:59Z CJP $
# Copyright:   (c) 2008 Rob McMullen
# License:     wxWidgets
#-----------------------------------------------------------------------------
#
# Originally based on code from Luke-SDK, which includes the following
# copyright notice:
# 
# Copyright (c) 2007 Eur Ing Christopher Thoday, cthoday@mail.berlios.de.
#
# Permission to use, copy, modify and distribute this software and its
# documentation for any purpose and without fee is hereby granted, provided
# that the above copyright notice appear in all copies and that both
# the copyright notice and this permission notice appear in supporting
# documentation.
#
# Christopher Thoday makes no representations about the suitability of this
# software for any purpose.  It is provided "as is" without express or implied
# warranty.

"""Spell checking for the wx.StyledTextControl using pyenchant

This module was insipred by the spell check function from Christopher Thoday's
U{Luke SDK<http://luke-sdk.berlios.de/>}.

Spell checking is provided by the pyenchant library, which is an external
dependency not part of wxPython.  Packages are available for Mac, Unix, and
windows at U{http://pyenchant.sourceforge.net}

Currently provides:
 - spell checking of entire buffer, currently visible page, or selection region
 - user specified indicator number (0 - 2), style, and color
 - language can be changed on the fly
 - update the spelling as you type
 - check the document in either idle time or in a background thread

@author: Rob McMullen
@version: 1.2

Changelog::
    1.2:
        - Rewrote as a standalone class rather than a static mixin
    1.1:
        - Added helper function to use idle processing time to check document
        - Added word checking function for use in instant spell checking
    1.0:
        - First public release
"""

import os
import locale
import wx
import wx.stc

# Assume MacPorts install of Enchant
if wx.Platform == '__WXMAC__':
    if 'PYENCHANT_LIBRARY_PATH' not in os.environ:
        os.environ['PYENCHANT_LIBRARY_PATH'] = '/opt/local/lib/libenchant.dylib'

try:
    import enchant
except ImportError:
    # no big deal; support for enchant simply won't be included
    pass
except:
    # big deal; enchant is there but there's some error that is preventing
    # its import
    import traceback
    traceback.print_exc()

class STCSpellCheck(object):
    """Spell checking for use with wx.StyledTextControl.
    
    This shows spelling errors using the styling indicators (e.g.  the red
    squiggly underline) of the styled text control; I find this much more
    convenient than a dialog-box that makes you click through each mistake.
    
    The eventual goal of the module is to provide on-the-fly spell checking
    that will display errors as you type, and also will highlight errors
    during idle time or in a background thread.
    
    Spell checking is provided through the pyenchant module.  Without
    pyenchant, this object won't do anything useful, but it is still safe to
    be used.  It wraps all calls to pyenchant with try/except blocks to catch
    import errors, and any calls to the spell checking functions will return
    immediately.
    
    To use the spelling check, use one of the methods L{checkAll},
    L{checkCurrentPage}, or L{checkSelection}.  Clear the spelling
    indicators with L{clearAll}.
    """
    # Class attributes to act as default values
    _spelling_lang = None
    _spelling_dict = None
    
    def __init__(self, stc, *args, **kwargs):
        """Mixin must be initialized using this constructor.
        
        Keyword arguments are also available instead of calling the
        convenience functions.  For L{setIndicator}, use C{indicator},
        C{indicator_color}, and {indicator_style}; for L{setLanguage},
        use C{language}; and for L{setMinimumWordSize}, use
        C{min_word_size}.  See the descriptions of those methods for more info.
        """
        self.stc = stc
        self.setIndicator(kwargs.get('indicator', 2),
                          kwargs.get('indicator_color', "#FF0000"),
                          kwargs.get('indicator_style', wx.stc.STC_INDIC_SQUIGGLE))
        self.setMinimumWordSize(kwargs.get('min_word_size', 3))
        if 'language' in kwargs:
            # Don't set default language unless explicitly specified -- it
            # might have already been set through the class method
            self.setDefaultLanguage(kwargs['language'])
        if 'check_region' in kwargs:
            # optional function to specify if the region should be spell
            # checked.  Function should return True if the position should
            # be spell-checked; False if it doesn't make sense to spell check
            # that part of the document
            self._spell_check_region = kwargs['check_region']
        else:
            self._spell_check_region = lambda s: True
        self._spelling_debug = False
        
        self._spelling_last_idle_line = -1
        self.dirty_range_count_per_idle = 5
        
        self._no_update = False
        self._last_block = -1
        
        self.clearDirtyRanges()

    def setIndicator(self, indicator=None, color=None, style=None):
        """Set the indicator styling for misspelled words.
        
        Set the indicator index to use, its color, and the visual style.
        
        @param indicator: indicator number (usually 0, 1, or 2, but may be fewer
        depending on the number of style bits you've chosen for the stc.)
        @param color: string indicating the color of the indicator (e.g.
        "#FF0000" for red)
        @param style: stc indicator style; one of the wx.stc.STC_INDIC_*
        constants (currently wx.stc.STC_INDIC_PLAIN, wx.stc.STC_INDIC_SQUIGGLE,
        wx.stc.STC_INDIC_TT, wx.stc.STC_INDIC_DIAGONAL,
        wx.stc.STC_INDIC_STRIKE, wx.stc.STC_INDIC_HIDDEN,
        wx.stc.STC_INDIC_BOX, wx.stc.STC_INDIC_ROUNDBOX)
        """
        indicators = {0: wx.stc.STC_INDIC0_MASK,
                      1: wx.stc.STC_INDIC1_MASK,
                      2: wx.stc.STC_INDIC2_MASK
                      }
        if indicator is not None:
            if indicator not in indicators:
                indicator = 0
            # The current view may have fewer than 3 indicators
            bitmax = 7 - self.stc.GetStyleBits()
            if indicator > bitmax:
                indicator = bitmax
            self._spelling_indicator = indicator
        self._spelling_indicator_mask = indicators[self._spelling_indicator]
        
        if color is not None:
            self._spelling_color = color
        self.stc.IndicatorSetForeground(self._spelling_indicator,
                                    self._spelling_color)
    
        if style is not None:
            if style > wx.stc.STC_INDIC_MAX:
                style = wx.stc.STC_INDIC_MAX
            self._spelling_style = style
        self.stc.IndicatorSetStyle(self._spelling_indicator,
                               self._spelling_style)
    
    @classmethod
    def getAvailableLanguages(cls):
        """Return a list of supported languages.
        
        Pyenchant supplies a list of its supported languages, so this is just
        a simple wrapper around its C{list_languages} function.  Each item in
        the list is a text string indicating the locale name, e.g.  en_US, ru,
        ru_RU, eo, es_ES, etc.
        
        @return: a list of text strings indicating the supported languages
        """
        try:
            return enchant.list_languages()
        except NameError:
            pass
        return []
    
    @classmethod
    def _getDict(cls, lang):
        try:
            d = enchant.Dict(lang)
        except:
            # Catch all exceptions, because if pyenchant isn't available, you
            # can't catch the enchant.DictNotFound error.
            d = None
        return d

    def setCheckRegion(self, func):
        """Set region checker callable
        @param func: def func(pos): return bool

        """
        self.clearAll()
        self._spell_check_region = func

    @classmethod
    def setDefaultLanguage(cls, lang):
        """Set the default language for spelling check.
        
        The string should be in language locale format, e.g.  en_US, ru, ru_RU,
        eo, es_ES, etc.  See L{getAvailableLanguages}.
        
        @param lang: text string indicating the language
        """
        cls._spelling_lang = lang
        cls._spelling_dict = cls._getDict(lang)

    @classmethod
    def getSpellingDictionary(cls):
        """Get the currently used spelling dictionary
        @return: enchant.Dict instance or None

        """
        return cls._spelling_dict

    def setLanguage(self, lang):
        """Set the language for spelling check for this class, if different than
        the default.
        
        The string should be in language locale format, e.g.  en_US, ru, ru_RU,
        eo, es_ES, etc.  See L{getAvailableLanguages}.
        
        @param lang: text string indicating the language
        """
        # Note that this instance variable will shadow the class attribute
        self._spelling_lang = lang
        self._spelling_dict = self._getDict(lang)
    
    def hasDictionary(self):
        """Returns True if a dictionary is available to spell check the current
        language.
        """
        return self._spelling_dict is not None

    @classmethod
    def isEnchantOk(cls):
        """Returns True if enchant is available"""
        return 'enchant' in globals()

    @classmethod
    def reloadEnchant(cls, libpath=u''):
        """Try (re)loading the enchant module. Use to dynamically try to
        import enchant incase it could be loaded at the time of the import of
        this module.
        @keyword libpath: optionally specify path to libenchant
        @return: bool

        """
        try:
            if libpath and os.path.exists(libpath):
                os.environ['PYENCHANT_LIBRARY_PATH'] = libpath

            if cls.isEnchantOk():
                reload(enchant)
            else:
                mod = __import__('enchant', globals(), locals())
                globals()['enchant'] = mod
        except ImportError:
            return False
        else:
            return True

    def getLanguage(self):
        """Returns True if a dictionary is available to spell check the current
        language.
        """
        return self._spelling_lang
    
    def setMinimumWordSize(self, size):
        """Set the minimum word size that will be looked up in the dictionary.
        
        Words smaller than this size won't be spell checked.
        """
        self._spelling_word_size = size
    
    def clearAll(self):
        """Clear the stc of all spelling indicators."""
        self.stc.StartStyling(0, self._spelling_indicator_mask)
        self.stc.SetStyling(self.stc.GetLength(), 0)
    
    def checkRange(self, start, end):
        """Perform a spell check over a range of text in the document.
        
        This is the main spell checking routine -- it loops over the range
        of text using the L{findNextWord} method to break the text into
        words to check.  Misspelled words are highlighted using the current
        indicator.
        
        @param start: starting position
        @param end: last position to check
        """
        spell = self._spelling_dict
        if not spell:
            return
        
        # Remove any old spelling indicators
        mask = self._spelling_indicator_mask
        count = end - start
        if count <= 0:
            if self._spelling_debug:
                print("No need to check range: start=%d end=%d count=%d" % (start, end, count))
            return
        self.stc.StartStyling(start, mask)
        self.stc.SetStyling(count, 0)
        
        text = self.stc.GetTextRange(start, end) # note: returns unicode
        unicode_index = 0
        max_index = len(text)
        
        last_index = 0 # last character in text a valid raw byte position
        last_pos = start # raw byte position corresponding to last_index
        while unicode_index < max_index:
            start_index, end_index = self.findNextWord(text, unicode_index, max_index)
            if end_index >= 0:
                if end_index - start_index >= self._spelling_word_size:
                    if self._spelling_debug:
                        print("checking %s at text[%d:%d]" % (repr(text[start_index:end_index]), start_index, end_index))
                    if not spell.check(text[start_index:end_index]):
                        # Because unicode characters are stored as utf-8 in the
                        # stc and the positions in the stc correspond to the
                        # raw bytes, not the number of unicode characters, we
                        # have to find out the offset to the unicode chars in
                        # terms of raw bytes.
                        
                        # find the number of raw bytes from the last calculated
                        # styling position to the start of the word
                        last_pos += len(text[last_index:start_index].encode('utf-8'))
                        
                        # find the length of the word in raw bytes
                        raw_count = len(text[start_index:end_index].encode('utf-8'))
                        
                        if self._spell_check_region(last_pos):
                            if self._spelling_debug:
                                print("styling text[%d:%d] = (%d,%d) to %d" % (start_index, end_index, last_pos, last_pos + raw_count, mask))
                            self.stc.StartStyling(last_pos, mask)
                            self.stc.SetStyling(raw_count, mask)
                        elif self._spelling_debug:
                            print("not in valid spell check region.  styling position corresponding to text[%d:%d] = (%d,%d)" % (start_index, end_index, last_pos, last_pos + raw_count))
                        last_pos += raw_count
                        last_index = end_index
                unicode_index = end_index
            else:
                break

    def checkAll(self):
        """Perform a spell check on the entire document."""
        return self.checkRange(0, self.stc.GetLength())
    
    def checkSelection(self):
        """Perform a spell check on the currently selected region."""
        return self.checkRange(self.stc.GetSelectionStart(), self.stc.GetSelectionEnd())
    
    def checkLines(self, startline=-1, count=-1):
        """Perform a spell check on group of lines.
        
        Given the starting line, check the spelling on a block of lines.  If
        the number of lines in the block is not specified, use the number of
        currently visibile lines.
        
        @param startline: current line, or -1 to use the first visible line
        @param count: number of lines in the block, or -1 to use the number of
        lines visible on screen
        """
        if startline < 0:
            startline = self.stc.GetFirstVisibleLine()
        start = self.stc.PositionFromLine(startline)
        if count < 0:
            count = self.stc.LinesOnScreen()
        endline = startline + count
        if endline > self.stc.GetLineCount():
            endline = self.stc.GetLineCount() - 1
        end = self.stc.GetLineEndPosition(endline)
        if self._spelling_debug:
            print("Checking lines %d-%d, chars %d=%d" % (startline, endline, start, end))
        return self.checkRange(start, end)
    
    def checkCurrentPage(self):
        """Perform a spell check on the currently visible lines."""
        return self.checkLines()
    
    def findNextWord(self, utext, index, length):
        """Find the next valid word to check.
        
        Designed to be overridden in subclasses, this method takes a starting
        position in an array of text and returns a tuple indicating the next
        valid word in the string.
        
        @param utext: array of unicode chars
        @param i: starting index within the array to search
        @param length: length of the text
        @return: tuple indicating the word start and end indexes, or (-1, -1)
        indicating that the end of the array was reached and no word was found
        """
        while index < length:
            if utext[index].isalpha():
                end = index + 1
                while end < length and utext[end].isalpha():
                    end += 1
                return (index, end)
            index += 1
        return (-1, -1)
    
    def startIdleProcessing(self):
        """Initialize parameters needed for idle block spell checking.
        
        This must be called before the first call to L{processIdleBlock}
        or if you wish to restart the spell checking from the start
        of the document.  It initializes parameters needed by the
        L{processIdleBlock} in order to process the document during idle
        time.
        """
        self._spelling_last_idle_line = 0
        
    def processIdleBlock(self):
        """Process a block of lines during idle time.
        
        This method is designed to be called during idle processing and will
        spell check a small number of lines.  The next idle processing event
        will continue from where the previous call left off, and in this way
        over some number of idle events will spell check the entire document.
        
        Once the entire document is spell checked, a flag is set and
        further calls to this method will immediately return.  Calling
        L{startIdleProcessing} will cause the idle processing to start
        checking from the beginning of the document.
        """
        self.processDirtyRanges()
        if self._spelling_last_idle_line < 0:
            return
        if self._spelling_debug:
            print("Idle processing page starting at line %d" % self._spelling_last_idle_line)
        self.checkLines(self._spelling_last_idle_line)
        self._spelling_last_idle_line += self.stc.LinesOnScreen()
        if self._spelling_last_idle_line > self.stc.GetLineCount():
            self._spelling_last_idle_line = -1
            return False
        return True

    def processCurrentlyVisibleBlock(self):
        """Alternate method to check lines during idle time.
        
        This method is designed to be called during idle processing and will
        spell check the currently visible block of lines.  Once the visible
        block has been checked, repeatedly calling this method will have
        no effect until the line position changes (or in the less frequent
        occurrence when the number of lines on screen changes by resizing
        the window).
        """
        self.processDirtyRanges()

        self._spelling_last_idle_line = self.stc.GetFirstVisibleLine()
        curr_block = self._spelling_last_idle_line + self.stc.LinesOnScreen()
        if self._no_update or curr_block == self._last_block:
            return

        self.checkLines(self._spelling_last_idle_line)
        self._spelling_last_idle_line += self.stc.LinesOnScreen()
        self._last_block = self._spelling_last_idle_line
        return True

    def getSuggestions(self, word):
        """Get suggestion for the correct spelling of a word.
        
        @param word: word to check
        
        @return: list of suggestions, or an empty list if any of the following
        are true: there are no suggestions, the word is shorter than the
        minimum length, or the dictionary can't be found.
        """
        spell = self._spelling_dict
        if spell and len(word) >= self._spelling_word_size:
            words = spell.suggest(word)
            if self._spelling_debug:
                print("suggestions for %s: %s" % (word, words))
            return words
        return []
    
    def checkWord(self, pos=None, atend=False):
        """Check the word at the current or specified position.
        
        @param pos: position of a character in the word (or at the start or end
        of the word), or None to use the current position
        @param atend: True if you know the cursor is at the end of the word
        """
        if pos is None:
            pos = self.stc.GetCurrentPos()
        if atend:
            end = pos
        else:
            end = self.stc.WordEndPosition(pos, True)
        start = self.stc.WordStartPosition(pos, True)
        if self._spelling_debug:
            print("%d-%d: %s" % (start, end, self.stc.GetTextRange(start, end)))
        self.checkRange(start, end)
    
    def addDirtyRange(self, start, end, lines_added=0, deleted=False):
        """Add a range of characters to a list of dirty regions that need to be
        updated when some idle time is available.
        
        """
        count = end - start
        if deleted:
            count = -count
        if start == self.current_dirty_end:
            self.current_dirty_end = end
        elif start >= self.current_dirty_start and start < self.current_dirty_end:
            self.current_dirty_end += count
        else:
            ranges = []
            if self.current_dirty_start >= 0:
                ranges.append((self.current_dirty_start, self.current_dirty_end))
            for range_start, range_end in self.dirty_ranges:
                if start < range_start:
                    range_start += count
                    range_end += count
                ranges.append((range_start, range_end))
            self.dirty_ranges = ranges
                
            self.current_dirty_start = start
            self.current_dirty_end = end
        
        # If there has been a change before the word that used to be under the
        # cursor, move the pointer so it matches the text
        if start < self.current_word_start:
            self.current_word_start += count
            self.current_word_end += count
        elif start <= self.current_word_end:
            self.current_word_end += count
            # Prevent nonsensical word end if lots of text have been deleted
            if self.current_word_end < self.current_word_start:
                #print("word start = %d, word end = %d" % (self.current_word_start, self.current_word_end))
                self.current_word_end = self.current_word_start
            
        if lines_added > 0:
            start = self.current_dirty_start
            line = self.stc.LineFromPosition(start)
            while True:
                line_end = self.stc.GetLineEndPosition(line)
                if line_end >= end:
                    #self.dirty_ranges.append((start, line_end))
                    if end > start:
                        self.current_dirty_start = start
                        self.current_dirty_end = end
                    else:
                        self.current_dirty_start = self.current_dirty_end = -1
                    break
                self.dirty_ranges.append((start, line_end))
                line += 1
                start = self.stc.PositionFromLine(line)
            
        if self._spelling_debug:
            print("event: %d-%d, current dirty range: %d-%d, older=%s" % (start, end, self.current_dirty_start, self.current_dirty_end, self.dirty_ranges))
    
    def clearDirtyRanges(self, ranges=None):
        """Throw away all dirty ranges
        
        """
        self.current_dirty_start = self.current_dirty_end = -1
        self.current_word_start = self.current_word_end = -1
        if ranges is not None:
            self.dirty_ranges = ranges
        else:
            self.dirty_ranges = []
    
    def processDirtyRanges(self):
        cursor = self.stc.GetCurrentPos()
        
        # Check that the cursor has moved off the current word and if so check
        # its spelling
        if self.current_word_start > 0:
            if cursor < self.current_word_start or cursor > self.current_word_end:
                self.checkRange(self.current_word_start, self.current_word_end)
                self.current_word_start = -1
        
        # Check spelling around the region currently being typed
        if self.current_dirty_start >= 0:
            range_start, range_end = self.processDirtyRange(self.current_dirty_start, self.current_dirty_end)
            
            # If the cursor is in the middle of a word, remove the spelling
            # markers
            if cursor >= range_start and cursor <= range_end:
                word_start = self.stc.WordStartPosition(cursor, True)
                word_end = self.stc.WordEndPosition(cursor, True)
                mask = self._spelling_indicator_mask
                self.stc.StartStyling(word_start, mask)
                self.stc.SetStyling(word_end - word_start, 0)
                
                if word_start != word_end:
                    self.current_word_start = word_start
                    self.current_word_end = word_end
                else:
                    self.current_word_start = -1
            self.current_dirty_start = self.current_dirty_end = -1
        
        # Process a chunk of dirty ranges
        needed = min(len(self.dirty_ranges), self.dirty_range_count_per_idle)
        ranges = self.dirty_ranges[0:needed]
        self.dirty_ranges = self.dirty_ranges[needed:]
        for start, end in ranges:
            if self._spelling_debug:
                print("processing %d-%d" % (start, end))
            self.processDirtyRange(start, end)
    
    def processDirtyRange(self, start, end):
        range_start = self.stc.WordStartPosition(start, True)
        range_end = self.stc.WordEndPosition(end, True)
        if self._spelling_debug:
            print("processing dirty range %d-%d (modified from %d-%d): %s" % (range_start, range_end, start, end, repr(self.stc.GetTextRange(range_start, range_end))))
        self.checkRange(range_start, range_end)
        return range_start, range_end


if __name__ == "__main__":
    import sys
    try:
        import enchant
    except:
        print("pyenchant not available, so spelling correction won't work.")
        print("Get pyenchant from http://pyenchant.sourceforge.net")
    
    class TestSTC(wx.stc.StyledTextCtrl):
        def __init__(self, *args, **kwargs):
            wx.stc.StyledTextCtrl.__init__(self, *args, **kwargs)
            self.spell = STCSpellCheck(self, language="en_US")
            self.SetMarginType(0, wx.stc.STC_MARGIN_NUMBER)
            self.SetMarginWidth(0, 32)
            self.Bind(wx.stc.EVT_STC_MODIFIED, self.OnModified)
            self.Bind(wx.EVT_IDLE, self.OnIdle)
            self.modified_count = 0
            self.idle_count = 0

        def OnModified(self, evt):
            # NOTE: on really big insertions, evt.GetText can cause a
            # MemoryError on MSW, so I've commented this dprint out.
            mod = evt.GetModificationType()
            if mod & wx.stc.STC_MOD_INSERTTEXT or mod & wx.stc.STC_MOD_DELETETEXT:
                #print("(%s) at %d: text=%s len=%d" % (self.transModType(evt.GetModificationType()),evt.GetPosition(), repr(evt.GetText()), evt.GetLength()))
                pos = evt.GetPosition()
                last = pos + evt.GetLength()
                self.spell.addDirtyRange(pos, last, evt.GetLinesAdded(), mod & wx.stc.STC_MOD_DELETETEXT)
                #self.modified_count += 1
                #if self.modified_count > 10:
                #    wx.CallAfter(self.spell.processDirtyRanges)
                #    self.modified_count = 0
            evt.Skip()
        
        def OnIdle(self, evt):
            #print("Idle")
            self.idle_count += 1
            if self.idle_count > 10:
                self.spell.processIdleBlock()
                self.idle_count = 0
            
        def transModType(self, modType):
            st = ""
            table = [(wx.stc.STC_MOD_INSERTTEXT, "InsertText"),
                     (wx.stc.STC_MOD_DELETETEXT, "DeleteText"),
                     (wx.stc.STC_MOD_CHANGESTYLE, "ChangeStyle"),
                     (wx.stc.STC_MOD_CHANGEFOLD, "ChangeFold"),
                     (wx.stc.STC_PERFORMED_USER, "UserFlag"),
                     (wx.stc.STC_PERFORMED_UNDO, "Undo"),
                     (wx.stc.STC_PERFORMED_REDO, "Redo"),
                     (wx.stc.STC_LASTSTEPINUNDOREDO, "Last-Undo/Redo"),
                     (wx.stc.STC_MOD_CHANGEMARKER, "ChangeMarker"),
                     (wx.stc.STC_MOD_BEFOREINSERT, "B4-Insert"),
                     (wx.stc.STC_MOD_BEFOREDELETE, "B4-Delete")
                     ]

            for flag,text in table:
                if flag & modType:
                    st = st + text + " "

            if not st:
                st = 'UNKNOWN'

            return st

    class Frame(wx.Frame):
        def __init__(self, *args, **kwargs):
            super(self.__class__, self).__init__(*args, **kwargs)

            self.stc = TestSTC(self, -1)

            self.CreateStatusBar()
            menubar = wx.MenuBar()
            self.SetMenuBar(menubar)  # Adding the MenuBar to the Frame content.
            menu = wx.Menu()
            menubar.Append(menu, "File")
            self.menuAdd(menu, "Open", "Open File", self.OnOpenFile)
            self.menuAdd(menu, "Quit", "Exit the pragram", self.OnQuit)
            menu = wx.Menu()
            menubar.Append(menu, "Edit")
            self.menuAdd(menu, "Check All", "Spell check the entire document", self.OnCheckAll)
            self.menuAdd(menu, "Check Current Page", "Spell check the currently visible page", self.OnCheckPage)
            self.menuAdd(menu, "Check Selection", "Spell check the selected region", self.OnCheckSelection)
            menu.AppendSeparator()
            self.menuAdd(menu, "Clear Spelling", "Remove spelling correction indicators", self.OnClearSpelling)
            menu = wx.Menu()
            menubar.Append(menu, "Language")
            langs = self.stc.spell.getAvailableLanguages()
            self.lang_id = {}
            for lang in langs:
                id = wx.NewId()
                self.lang_id[id] = lang
                self.menuAdd(menu, lang, "Change dictionary to %s" % lang, self.OnChangeLanguage, id=id)


        def loadFile(self, filename):
            fh = open(filename)
            self.stc.SetText(fh.read())
            self.stc.spell.clearDirtyRanges()
            self.stc.spell.checkCurrentPage()
        
        def loadSample(self, paragraphs=10):
            lorem_ipsum = u"""\
Lorem ipsum dolor sit amet, consectetuer adipiscing elit.  Vivamus mattis
commodo sem.  Phasellus scelerisque tellus id lorem.  Nulla facilisi.
Suspendisse potenti.  Fusce velit odio, scelerisque vel, consequat nec,
dapibus sit amet, tortor.  Vivamus eu turpis.  Nam eget dolor.  Integer
at elit.  Praesent mauris.  Nullam non nulla at nulla tincidunt malesuada.
Phasellus id ante.  Sed mauris.  Integer volutpat nisi non diam.  Etiam
elementum.  Pellentesque interdum justo eu risus.  Cum sociis natoque
penatibus et magnis dis parturient montes, nascetur ridiculus mus.  Nunc
semper.  In semper enim ut odio.  Nulla varius leo commodo elit.  Quisque
condimentum, nisl eget elementum laoreet, mauris turpis elementum felis, ut
accumsan nisl velit et mi.

And some Russian: \u041f\u0438\u0442\u043e\u043d - \u043b\u0443\u0447\u0448\u0438\u0439 \u044f\u0437\u044b\u043a \u043f\u0440\u043e\u0433\u0440\u0430\u043c\u043c\u0438\u0440\u043e\u0432\u0430\u043d\u0438\u044f!

"""
            self.stc.ClearAll()
            for i in range(paragraphs):
                self.stc.AppendText(lorem_ipsum)
            # Call the spell check after the text has had a chance to be
            # displayed and the window resized to the correct size.
            self.stc.spell.clearDirtyRanges()
            wx.CallAfter(self.stc.spell.checkCurrentPage)

        def menuAdd(self, menu, name, desc, fcn, id=-1, kind=wx.ITEM_NORMAL):
            if id == -1:
                id = wx.NewId()
            a = wx.MenuItem(menu, id, name, desc, kind)
            menu.AppendItem(a)
            wx.EVT_MENU(self, id, fcn)
            menu.SetHelpString(id, desc)
        
        def OnOpenFile(self, evt):
            dlg = wx.FileDialog(self, "Choose a text file",
                               defaultDir = "",
                               defaultFile = "",
                               wildcard = "*")
            if dlg.ShowModal() == wx.ID_OK:
                print("Opening %s" % dlg.GetPath())
                self.loadFile(dlg.GetPath())
            dlg.Destroy()
        
        def OnQuit(self, evt):
            self.Close(True)
        
        def OnCheckAll(self, evt):
            self.stc.spell.checkAll()
        
        def OnCheckPage(self, evt):
            self.stc.spell.checkCurrentPage()
        
        def OnCheckSelection(self, evt):
            self.stc.spell.checkSelection()
        
        def OnClearSpelling(self, evt):
            self.stc.spell.clearAll()
        
        def OnChangeLanguage(self, evt):
            id = evt.GetId()
            normalized = locale.normalize(self.lang_id[id])
            try:
                locale.setlocale(locale.LC_ALL, normalized)
                print("Changing locale %s, dictionary set to %s" % (normalized, self.lang_id[id]))
            except locale.Error:
                print("Can't set python locale to %s; dictionary set to %s" % (normalized, self.lang_id[id]))
            self.stc.spell.setLanguage(self.lang_id[id])
            self.stc.spell.clearAll()
            self.stc.spell.checkCurrentPage()

    app = wx.App(False)
    frame = Frame(None, size=(600, -1))
    need_sample = True
    if len(sys.argv) > 1:
        if not sys.argv[-1].startswith("-"):
            frame.loadFile(sys.argv[-1])
            need_sample = False
    if need_sample:
        frame.loadSample()
    if '-d' in sys.argv:
        frame.stc.spell._spelling_debug = True
    frame.Show()
    app.MainLoop()



###############################################################################
# Name: eiffel.py                                                             #
# Purpose: Define Eiffel syntax for highlighting and other features           #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: eiffel.py
AUTHOR: Cody Precord
@summary: Lexer configuration module for Eiffel
@todo: look into why io.anything is highlighted as a number

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _eiffel.py 68798 2011-08-20 17:17:05Z CJP $"
__revision__ = "$Revision: 68798 $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc

# Local Imports
import synglob
import syndata

#-----------------------------------------------------------------------------#

#---- Keyword Definitions ----#
EIFFEL_KW = (0, "alias all and any as bit boolean check class character clone "
                "cluster create creation current debug deferred div do double "
                "else elseif end ensure equal expanded export external false "
                "feature forget from frozen general if implies indexing infix "
                "inherit inspect integer invariant is language like local loop "
                "mod name nochange none not obsolete old once or platform "
                "pointer prefix precursor program real redefine rename require "
                "rescue result retry root select separate string strip then "
                "true undefine unique until variant void when xor")
#---- End Keyword Definitions ----#

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [(stc.STC_EIFFEL_CHARACTER, 'char_style'),
                (stc.STC_EIFFEL_COMMENTLINE, 'comment_style'),
                (stc.STC_EIFFEL_DEFAULT, 'default_style'),
                (stc.STC_EIFFEL_IDENTIFIER, 'default_style'),
                (stc.STC_EIFFEL_NUMBER, 'number_style'),
                (stc.STC_EIFFEL_OPERATOR, 'operator_style'),
                (stc.STC_EIFFEL_STRING, 'string_style'),
                (stc.STC_EIFFEL_STRINGEOL, 'stringeol_style'),
                (stc.STC_EIFFEL_WORD, 'keyword_style')]

#---- Extra Properties ----#
FOLD = ("fold", "1")

#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for Eiffel""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_EIFFEL)

    def GetKeywords(self):
        """Returns Specified Keywords List """
        return [EIFFEL_KW]

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return SYNTAX_ITEMS

    def GetProperties(self):
        """Returns a list of Extra Properties to set """
        return [FOLD]

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code """
        return [u'--']



###############################################################################
# Name: testSyntaxDataBase.py                                                 #
# Purpose: Unit tests for syntax.syndata Base Class                           #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2010 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""Unittest cases for testing syndata.SyntaxDataBase class"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: testSyntaxDataBase.py 63836 2010-04-03 06:16:55Z CJP $"
__revision__ = "$Revision: 63836 $"

#-----------------------------------------------------------------------------#
# Imports
import unittest
import wx
import wx.stc

# Module to test
import syntax.syndata as syndata

#-----------------------------------------------------------------------------#
# Test Class

class SyntaxDataBaseTest(unittest.TestCase):

    def setUp(self):
        self.data = syndata.SyntaxDataBase()

    def tearDown(self):
        pass

    #---- Test Cases ----#

    def testCommentPattern(self):
        self.assertTrue(isinstance(self.data.CommentPattern, list))
        self.assertTrue(self.data.CommentPattern == self.data.GetCommentPattern())

    def testKeywords(self):
        self.assertTrue(isinstance(self.data.Keywords, list))
        self.assertTrue(self.data.Keywords == self.data.GetKeywords())

    def testLangId(self):
        self.assertTrue(isinstance(self.data.LangId, int))
        self.assertTrue(self.data.LangId == self.data.GetLangId())

    def testLexer(self):
        self.assertTrue(isinstance(self.data.Lexer, int)) 
        lexers = [ getattr(wx.stc, item) for item in dir(wx.stc)
                   if item.startswith('STC_LEX') ]
        self.assertTrue(self.data.Lexer in lexers)

    def testProperties(self):
        self.assertTrue(isinstance(self.data.Properties, list))
        self.assertTrue(self.data.Properties == self.data.GetProperties())

    def testSyntaxSpec(self):
        self.assertRaises(NotImplementedError, self.data.GetSyntaxSpec)

    def testRegisterFeature(self):
        def foo():
            pass
        self.data.RegisterFeature('foo', foo)
        self.assertTrue(self.data.GetFeature('foo') is foo)

    def testSetLangId(self):
        self.data.SetLangId(10)
        self.assertEquals(10, self.data.LangId)

    def testSetLexer(self):
        self.data.SetLexer(wx.stc.STC_LEX_CPP)
        self.assertEquals(wx.stc.STC_LEX_CPP, self.data.GetLexer())



###############################################################################
# Name: pycomp.py                                                             #
# Purpose: Provides python autocompletion lists and calltips for the editor   #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
Provides completion and calltip support for python documents. To provide the
completion lists and calltips a mix of parsing and introspection is used to
deduct the requested information.

@summary: Python autocompletion support

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__cvsid__ = "$Id: pycomp.py 70229 2012-01-01 01:27:10Z CJP $"
__revision__ = "$Revision: 70229 $"

#--------------------------------------------------------------------------#
# Dependancies
import os
import sys
import time
import tokenize
import types
from token import NAME, DEDENT, STRING

import wx
from wx.py import introspect

from StringIO import StringIO

# Local imports
import completer

#--------------------------------------------------------------------------#

class Completer(completer.BaseCompleter):
    """Python Code completion provider"""
    def __init__(self, stc_buffer):
        """Initializes the completer
        @param stc_buffer: buffer that contains code

        """
        super(Completer, self).__init__(stc_buffer)

        # Setup
        self.SetAutoCompKeys([ord('.'), ])
        self.SetAutoCompStops(' \'"\\`):')
        self.SetAutoCompFillups('.,;([]}<>%^&+-=*/|')
        self.SetCallTipKeys([ord('('), ])
        self.SetCallTipCancel([ord(')'), wx.WXK_RETURN])

        # Needed for introspect to run
        try:
            sys.ps1
        except AttributeError:
            sys.ps1 = '>>>'

        try:
            sys.ps2
        except AttributeError:
            sys.ps2 = '...'

    def _GetCompletionInfo(self, command, calltip=False):
        """Get Completion list or Calltip
        @return: list or string

        """
        if command is None or (len(command) and command[0].isdigit()):
            if calltip:
                return u""
            return list()

        try:
            cmpl = PyCompleter()

            # Put the files directory on the path so eval has a better
            # chance of getting the proper completions
            fname = self._buffer.GetFileName()
            if fname:
                fpath = os.path.dirname(fname)
                sys.path.insert(0, fpath)
            snapshot = list(sys.modules.keys())

            t1 = time.time()
            cmpl.evalsource(self._buffer.GetText(),
                            self._buffer.GetCurrentLine())
            dbg("[pycomp][info] Completion eval time: %f" % (time.time() - t1))

            if fname:
                sys.path.pop(0)

            # Dump any other modules that got brought in during eval
            # so that they get properly updated on next pass through.
            nsnapshot = sys.modules.keys()
            nimport = list(set(nsnapshot).difference(set(snapshot)))
            for k in nimport:
                del sys.modules[k]

            if calltip:
                return cmpl.get_completions(command + u'(', u'', calltip)
            else:
                # Get Auto-completion List
                complst = cmpl.get_completions(command)
                sigs = list()
                tmap = {"function" : completer.TYPE_FUNCTION,
                        "method" : completer.TYPE_METHOD,
                        "class" : completer.TYPE_CLASS,
                        "attribute" : completer.TYPE_ATTRIBUTE,
                        "property" : completer.TYPE_PROPERTY}
                for sig in complst:
                    word = sig['word'].rstrip(u'(.')
                    tval = tmap.get(sig['type'], completer.TYPE_UNKNOWN)
                    sigs.append(completer.Symbol(word, tval))
                sigs.sort(key=lambda x: x.Name.upper())
                return sigs

        except BaseException, msg:
            self._log("[pycomp][err] _GetCompletionInfo: %s, %s" % \
                      (sys.exc_info()[0], sys.exc_info()[1]))
            if calltip:
                return u""
            else:
                return list()
        
    def GetAutoCompList(self, command):
        """Returns the list of possible completions for a command string.
        @param command: command lookup is done on

        """
        return self._GetCompletionInfo(command)

    def GetCallTip(self, command):
        """Returns the formatted calltip string for the command.
        @param command: command to get calltip for

        """
        # get the relevant text
        alltext = self._GetCompletionInfo(command, calltip=True)
       
        # split the text into natural paragraphs (a blank line separated)
        paratext = alltext.split("\n\n")
       
        # add text by paragraph until text limit or all paragraphs
        textlimit = 800
        if len(paratext[0]) < textlimit:
            numpara = len(paratext)
            calltiptext = paratext[0]
            ii = 1
            while ii < numpara and \
                  (len(calltiptext) + len(paratext[ii])) < textlimit:
                calltiptext = calltiptext + "\n\n" + paratext[ii]
                ii = ii + 1

            # if not all texts are added, add "[...]"
            if ii < numpara:
                calltiptext = calltiptext + "\n[...]"
        # present the function signature only (first newline)
        else:
            calltiptext = alltext.split("\n")[0]

        if type(calltiptext) != types.UnicodeType:
            # Ensure it is unicode
            try:
                stcbuff = self.GetBuffer()
                encoding = stcbuff.GetEncoding()
                calltiptext = calltiptext.decode(encoding)
            except Exception, msg:
                dbg("%s" % msg)

        return calltiptext

#-----------------------------------------------------------------------------#
# This code below is a modified and adapted version of the pythoncomplete 
# Omni completion script for vim. The original vimscript can be found at the 
# following address: http://www.vim.org/scripts/script.php?script_id=1542
def dbg(msg):
    wx.GetApp().GetLog()(msg)

class PyCompleter(object):
    """Python code completion provider"""
    def __init__(self):
        self.compldict = {}
        self.parser = PyParser()

    def evalsource(self, text, line=0):
        """Evaluate source for introspection
        @param text: Text to evaluate
        @keyword line: current line of cursor

        """
        scope = self.parser.parse(text.replace('\r\n', '\n'), line)
        src = scope.get_code()
        # Test
#        f = open('pycompout.py', 'w')
#        f.write(src)
#        f.close()
        # NOTE: keep commented out in production code can cause slow down
        #       when running in editra with DEBUG output.
#        dbg("[pycomp][info] Generated source: %s" % src)
        try: 
            exec src in self.compldict
        except Exception, msg:
            dbg("[pycomp][err] src exec: %s" % msg)
        else:
            dbg("[pycomp][info] Successfully executed source code")

        for loc in [ l[1] for l in scope.locals]:
            try: 
                exec loc in self.compldict
            except Exception, msg:
                dbg("[pycomp][err] local exec %s [%s]" % (msg, loc))
            else:
                dbg("[pycomp][info] Successfully executed: %s" % loc)

    def get_arguments(self, func_obj):
        """Get the arguments of a given function obj
        @param func_obj: function object to get parameters for

        """
        def _ctor(obj):
            """Get the constructor for an object"""
            try:
                return obj.__init__.im_func
            except AttributeError:
                for base in obj.__bases__:
                    constructor = getattr(base, '__init__', None)
                    if constructor is not None:
                        return constructor
            return None

        arg_offset = 1
        if type(func_obj) == types.ClassType:
            func_obj = _ctor(func_obj)
        elif type(func_obj) == types.MethodType:
            func_obj = func_obj.im_func
        else: 
            arg_offset = 0
        
        arg_text = ''
        if type(func_obj) in [types.FunctionType, types.LambdaType]:
            try:
                fcode = func_obj.func_code
                real_args = fcode.co_varnames[arg_offset:fcode.co_argcount]
                defaults = func_obj.func_defaults or ''
                defaults = [ "=%s" % name for name in defaults ]
                defaults = [""] * (len(real_args) - len(defaults)) + defaults
                items = [ arg + default 
                          for arg, default in zip(real_args, defaults) ]
                if fcode.co_flags & 0x4:
                    items.append("...")
                if fcode.co_flags & 0x8:
                    items.append("***")
                arg_text = (','.join(items)) + ')'

            except Exception, msg:
                dbg("[pycomp][err] get_arguments: %s" % msg)

        if len(arg_text) == 0:
            # The doc string sometimes contains the function signature
            #  this works for alot of C modules that are part of the
            #  standard library
            doc = getattr(func_obj, '__doc__', False)
            if doc:
                doc = str(doc)
                doc = doc.lstrip()
                pos = doc.find('\n')
                if pos > 0:
                    sigline = doc[:pos]
                    lidx = sigline.find('(')
                    ridx = sigline.find(')')
                    if lidx > 0 and ridx > 0:
                        arg_text = sigline[lidx + 1 : ridx] + ')'

        if len(arg_text) == 0:
            arg_text = ')'

        return arg_text

    def get_completions(self, context, match='', ctip=False):
        """Get the completions for the given context
        @param context: command string to get completions for
        @keyword match: for matching an incomplete command string
        @keyword ctip: Get a calltip for the context instead of completion list
        @return: list of dictionaries

        """
        dbg("[pycomp] get_completions('%s','%s')" % (context, match))
        stmt = context + match

        try:
            result = None
            compdict = {}
            ridx = stmt.rfind('.')
            if len(stmt) > 0 and stmt[-1] == '(':
                if ctip:
                    # Try introspect.getCallTip since it is generally
                    # better at getting tips for c modules
                    tip = introspect.getCallTip(_sanitize(stmt), 
                                                self.compldict)[2]
                    if not isinstance(tip, basestring):
                        tip = u""
                    if not tip:
                        # Internal calltip code
                        result = eval(_sanitize(stmt.rstrip('(')), self.compldict)
                        doc = max(getattr(result, '__doc__', ''), ' ')
                        argval = context + _cleanstr(self.get_arguments(result))
                        tip = '\n'.join([argval, _cleanstr(doc)])
                        dbg("[pycomp][info] Used internal calltips")
                    return tip

            elif ridx == -1:
                match = stmt
                compdict = self.compldict
            else:
                match = stmt[ridx+1:]
                stmt = _sanitize(stmt[:ridx])
                result = eval(stmt, self.compldict)
                compdict = dir(result)

            dbg("[pycomp] completing: stmt:%s" % stmt)
            completions = []
            isdict = isinstance(compdict, dict)
            for meth in compdict:
                if meth == "_PyCmplNoType":
                    continue #this is internal

                try:
#                     dbg('[pycomp] possible completion: %s' % meth)
                    if meth.find(match) == 0:
                        if result is None:
                            # NOTE: when result is none compdict is a list
                            inst = meth #compdict[meth]
                        else:
                            inst = getattr(result, meth, None)

                        # TODO: necessary check to handle some odd swig related
                        #       errors. Find out why type 'swigvarlink' causes
                        #       the exception Unknown C global variable.
                        if len(dir(inst)):
                            doc = getattr(inst, '__doc__', None)
                            if doc is None:
                                doc = max(getattr(result, '__doc__', ' '), ' ')
                        else:
                            doc = ' '

                        if isdict:
                            typestr = str(compdict[inst])
                        else:
                            typestr = str(inst)

                        comp = {'word' : meth,
                                'abbr' : meth,
                                'info' : _cleanstr(str(doc)),
                                'type' : typestr}

                        if "function" in typestr:
                            comp['word'] += '('
                            comp['abbr'] += '(' + _cleanstr(self.get_arguments(inst))
                            comp['type'] = "function"
                        elif "method" in typestr or "slot wrapper" in typestr:
                            comp['word'] += '('
                            comp['abbr'] += '(' + _cleanstr(self.get_arguments(inst))
                            comp['type'] = "method"
                        elif "module" in typestr:
                            comp['word'] += '.'
                            comp['type'] = "module"
                        elif "class" in typestr:
                            comp['word'] += '('
                            comp['abbr'] += '('
                            comp['type'] = "class"
                        elif "attribute" in typestr or \
                             (not typestr.startswith('__') and \
                              not typestr.startswith('<')):
                            comp['type'] = "attribute"
                        elif "property" in typestr: 
                            comp['type'] = "property"
#                        else:
#                            print typestr, meth

                        completions.append(comp)
                except Exception, msg:
                    dbg("[pycomp][err] inner completion: %s [stmt='%s']:" % (msg, stmt))

            return completions
        except Exception, msg:
            dbg("[pycomp][err] get_completions: %s [stmt='%s']" % (msg, stmt))
            if ctip:
                return u""
            return list()

#-----------------------------------------------------------------------------#
# Code objects

class Scope(object):
    """Base class for representing code objects"""
    def __init__(self, name, indent):
        """Initialize the scope
        @param name: name of this object
        @param indent: the indentation/level of this scope

        """
        super(Scope, self).__init__()

        # Attributes
        self.subscopes = list()
        self.docstr = u''
        self.locals = list()
        self.parent = None
        self.name = name
        self.indent = indent
        self.objid = -1         # Tracks order of declaration

    DocStr = property(lambda self: self.docstr,
                      lambda self, dstr: setattr(self, 'docstr', dstr))
    Locals = property(lambda self: self.locals,
                      lambda self, loc: setattr(self, 'locals', loc))
    Parent = property(lambda self: self.parent,
                      lambda self, parent: setattr(self, 'parent', parent))

    def Clone(self, indent=0):
        """Clone this scope object"""
        obj = Scope(self.name, indent)
        obj.DocStr = self.DocStr
        obj.Locals = list(self.Locals)
        obj.Parent = self.Parent
        for scope in self.subscopes:
            obj.subscopes.append((scope[0], scope[1].Clone(indent + 1)))
        obj.objid = self.objid
        return obj

    def NextObjId(self):
        self.objid += 1
        return self.objid

    def add(self, sub):
        """Push a subscope into this scope
        @param sub: sub scope to push

        """
        sub.parent = self
        self.subscopes.append((self.NextObjId(), sub))
        return sub

    def doc(self, docstr):
        """Format and set the doc string for this scope
        @param docstr: Docstring to format and set

        """
        dstr = docstr
        if len(dstr):
            while len(dstr) and dstr[0] in '"\' ':
                dstr = dstr[1:]

            while len(dstr) and dstr[-1] in '"\' ':
                dstr = dstr[:-1]
        dstr = '\n'.join([d.lstrip() for d in dstr.split('\n')])
        self.docstr = dstr.rstrip()

    def local(self, loc):
        """Add an object to the scopes locals
        @param loc: local object to add to locals

        """
        self._checkexisting(loc)
        self.locals.append((self.NextObjId(), loc))

    def _checkexisting(self, test):
        """Convenience function... keep out duplicates
        @param test: assignment statement to check for existence of
                     variable in the scopes locals

        """
        if '=' in test:
            var = test.split('=')[0].strip()
            for loc in self.locals:
                if '=' in loc[1] and var == loc[1].split('=')[0].strip():
                    self.locals.remove(loc)

    def get_code(self):
        """Get a string of code that represents this scope
        @return: string

        """
        cstr = '"""' + self.docstr + '"""\n'
        nonimport = list()
        for loc in self.locals:
            if loc[1].startswith('import') or loc[1].startswith('from'):
                cstr += ("try:\n    %s\nexcept ImportError:\n    pass\n" % loc[1])
            else:
                nonimport.append(loc)

        # we need to start with this, to fix up broken completions
        # hopefully this name is unique enough...
        cstr += 'class _PyCmplNoType:\n    def __getattr__(self,name):\n        return None\n'

        decls = self.subscopes + nonimport
        decls.sort(key=lambda x: x[0])
        for decl in [d[1] for d in decls]:
            if isinstance(decl, Scope):
                cstr += decl.get_code()
            else:
                cstr += decl + '\n'

        return cstr

    def pop(self, indent):
        """Pop the scope until it is at the level of the given
        indent.
        @param indent: indent level to pop scope to
        @return: scope of given indent level

        """
        outer = self
        while outer.parent != None and outer.indent >= indent:
            outer = outer.parent
        return outer

    def currentindent(self):
        """Return string of current scopes indent level
        @return: string of spaces

        """
        return '    ' * self.indent

    def childindent(self):
        """Return string the next scopes indentation level
        @return: string of spaces

        """
        return '    ' * (self.indent + 1)

class Class(Scope):
    """Class for representing a python class object for the parser"""
    def __init__(self, name, supers, indent):
        """initialize the class object
        @param name: name of class
        @param supers: classes super classes
        @param indent: scope of indentation

        """
        super(Class, self).__init__(name, indent)
        self.supers = supers

    def Clone(self, indent=0):
        """Create a clone of this object"""
        obj = Class(self.name, self.supers, indent)
        obj.DocStr = self.DocStr
        obj.Locals = list(self.Locals)
        obj.Parent = self.Parent
        obj.objid = self.objid
        for scope in self.subscopes:
            obj.subscopes.append((scope[0], scope[1].Clone(indent + 1)))
        return obj

    def get_code(self):
        """Get the code string representation of the Class object
        @return: string

        """
        cstr = '%sclass %s' % (self.currentindent(), self.name)
        if len(self.supers) > 0:
            cstr += '(%s)' % ','.join(self.supers)

        cstr += ':\n'
        if len(self.docstr) > 0:
            cstr += self.childindent() + '"""' + self.docstr + '"""\n'
        need_pass = True
        decls = self.locals + self.subscopes
        decls.sort(key=lambda x: x[0])
        for decl in [d[1] for d in decls]:
            need_pass = False
            if isinstance(decl, Scope):
                cstr += decl.get_code()
            else:
                cstr += ('%s%s\n' % (self.childindent(), decl))

        if need_pass:
            cstr += '%spass\n' % self.childindent()
        return cstr

    def local(self, loc):
        """Add an object to the scopes locals
        @param loc: local object to add to locals

        """
        if loc and '.' not in loc:
            super(Class, self).local(loc)

class Function(Scope):
    """Create a function object for representing a python function
    definition in the parser.

    """
    def __init__(self, name, params, indent):
        """Create the function object
        @param name: name of function
        @param params: the functions parameters
        @param indent: indentation level of functions declaration (scope)

        """
        super(Function, self).__init__(name, indent)
        self.params = params

    def Clone(self, indent=0):
        """Create a copy of the functions declaration at the given
        scope of indentation.
        @keyword indent: indentation level of the declaration

        """
        obj = Function(self.name, self.params, indent)
        obj.DocStr = self.DocStr
        obj.Locals = list(self.Locals)
        obj.Parent = self.Parent
        for scope in self.subscopes:
            obj.subscopes.append((scope[0], scope[1].Clone(indent + 1)))
        obj.objid = self.objid
        return obj

    def get_code(self):
        """Get code string representation of the function object
        @return: string

        """
        cstr = "%sdef %s(%s):\n" % \
            (self.currentindent(), self.name, ','.join(self.params))
        if len(self.docstr) > 0:
            cstr += self.childindent() + '"""' + self.docstr + '"""\n'
        cstr += "%spass\n" % self.childindent()
        return cstr

#-----------------------------------------------------------------------------#
# Main Parser

class PyParser(object):
    """Python parsing class"""
    def __init__(self):
        """Initialize and create the PyParser"""
        super(PyParser, self).__init__()

        # Attributes
        self.top = Scope('global', 0)
        self.scope = self.top
        self.gen = None  # Token Generator
        self.curline = 0 # Current parse line
        self.currentscope = None # Current scope in parse

    def _parsedotname(self, pre=None):
        """Parse a dotted name string
        @return: tuple of (dottedname, nexttoken)

        """
        name = list()
        if pre is None:
            tokentype, token = self.next()[:2]
            if token == '(':
                self._parenparse()
            if tokentype != NAME and token != '*':
                return ('', token)
        else:
            token = pre

        name.append(token)
        while True:
            tokentype, token = self.next()[:2]
            if token == '(':
                # Handle 'self.foo(a,b,c)'
                self._parenparse()
                break

            if token != '.':
                break

            tokentype, token = self.next()[:2]
            if tokentype != NAME:
                break
            name.append(token)
        return (".".join(name), token)

    def _parseimportlist(self):
        """Parse and collect import statements
        @return: list of imports

        """
        imports = []
        while True:
            name, token = self._parsedotname()
            if not name:
                break

            name2 = ''
            if token == 'as':
                name2, token = self._parsedotname()

            imports.append((name, name2))
            while token != "," and "\n" not in token:
                token = self.next()[1]

            if token != ",":
                break

        return imports

    def _parenparse(self):
        """Parse paren enclosed statement
        @return: list of named items enclosed in the parens

        """
        name = ''
        names = []
        level = 1
        while True:
            token = self.next()[1]
            if token in (')', ',') and level == 1:
                # Remove keyword assignments as they can
                # cause eval breakage when using undefined
                # vars.
                if '=' in name:
                    name = name.split('=')[0].strip()
                names.append(name)
                name = ''

            if token == '(':
                level += 1
            elif token == ')':
                level -= 1
                if level == 0:
                    break
            elif token == ',' and level == 1:
                pass
            else:
                name += str(token)

        return names

    def _parsefunction(self, indent):
        """Parse a function definition at the given scope of
        indentation and create a class token object from the
        results.
        @param indent: scope of functions declaration

        """
        self.scope = self.scope.pop(indent)
        tokentype, fname, findent = self.next()
        if tokentype != NAME:
            return None

        tokentype, open_paren, tindent = self.next()
        if open_paren != '(':
            return None

        params = self._parenparse()
        tokentype, colon, tindent = self.next()
        if colon != ':':
            return None

        return Function(fname, params, indent)

    def _parseclass(self, indent):
        """Parse a class definition at the given scope of
        indentation and create a class token object from the
        results.
        @param indent: scope of classes declaration

        """
        self.scope = self.scope.pop(indent)
        tokentype, cname = self.next()[:-1]
        if tokentype != NAME:
            return None

        super_cls = []
        tokentype, next = self.next()[:-1]
        if next == '(':
            super_cls = self._parenparse()
        elif next != ':':
            return None

        return Class(cname, super_cls, indent)

    def _parseassignment(self):
        """Parse a variable assignment to resolve the variables type
        for introspection.
        @return: string of objects type

        """
        assign = ''
        tokentype, token = self.next()[:-1]
        tokens = { 
                   # Dict
                   '{' : '{}', 'dict' : '{}',
                   'locals' : '{}', 'globals' : '{}',
                   # File
                   'open' : 'file', 'file' : 'file',
                   # List
                   '[' : '[]', 'list' : '[]',
                   'dir' : '["",""]', 'zip' : '[]', 'map' : '[]',
                   'sorted' : '[]', 'range' : '[]',
                   # NoneType
                   'None' : '_PyCmplNoType()',
                   # Number
                   tokenize.NUMBER : '0', 'ord' : '0', 'id' : '0',
                   'abs' : '0', 'sum' : '0', 'pow' : '0', 'len' : '0',
                   'hash' : '0',
                   # Property
                   'property' : 'property()',
                   # String
                   tokenize.STRING : '""', 'str' : '""',
                   'repr' : '""', 'chr' : '""', 'unichr' : '""',
                   'hex' : '""', 'oct' : '""', "'" : '""', '"' : '""',
                   # Type
                   'type' : 'type(_PyCmplNoType)',
                   # Tuple
                   '(' : '()', 'tuple' : '()', 'coerce' : '()'
                 }

        if tokentype == tokenize.NUMBER or tokentype == tokenize.STRING:
            return token #token = tokentype

        if token in tokens:
            return tokens[token]
        else:
            # NOTE: This part of the parse is where the problem is happening
            assign += token
            level = 0
            while True:
                tokentype, token = self.next()[:-1]
                if token in ('(', '{', '['):
                    level = level + 1
                elif token in (']', '}', ')'):
                    level = level - 1
                    if level == 0 or level == -1:
                        break
                elif level == 0:
                    # end of line
                    if token in (';', '\n', '='):
                        break
                    elif token in ('+', '*'):
                        # Attempt simple type deduction based on 
                        # operator statement.
                        if assign.endswith('"') or assign.endswith("'"):
                            assign = '""'
                        else:
                            assign = '0'
                        break
                    elif token in ('/', '-'):
                        # assignment most likely returns an digit type
                        # this is the best guess we can make
                        assign = '0'
                        break
                    elif token in ('and', 'or', 'in', '==',
                                   '<', '>', '!=', 'not', '>=', '<='):
                        # right side of assignment likely evaluates to bool
                        assign = 'bool'
                        break
#                    elif token == ',':
#                        assign = ''
                    else:
                        if token == '.' or assign.endswith('.'):
                            assign += token
                        else:
                            assign += (" %s" % token)

        # Check syntax to filter out bad tokens
        try:
            compile(assign, '_pycomp', 'eval')
        except:
            dbg("[pycomp][err] parseassignment bad token: %s" % assign)
            return '_PyCmplNoType()'
        else:
            return assign

    def next(self):
        """Get tokens of next line in parse
        @return: tuple of (type, token, indent)

        """
        ttype, token, (lineno, indent) = self.gen.next()[:3]
        if lineno == self.curline:
            self.currentscope = self.scope
        return (ttype, token, indent)

    def _adjustvisibility(self):
        """Adjust the visibility of the current contexts scope
        @return: current scope

        """
        newscope = Scope('result', 0)
        scopes = list()
        tscp = self.currentscope.parent
        while tscp != None:
            scopes.append(tscp)
            tscp = tscp.parent
        scopes.append(self.currentscope)

        for scp in scopes:
            if type(scp) == Function:
                cut = 0
                # Handle 'self' params
                if scp.parent != None and type(scp.parent) == Class:
                    cut = 1
                    params = scp.params[0]
                    ind = params.find('=')
                    if ind != -1:
                        params = params[:ind]
                    newscope.local('%s = %s' % (_sanitizeParam(scp.params[0]), scp.parent.name))

                for param in scp.params[cut:]:
                    ind = param.find('=')
                    if len(param) == 0:
                        continue
                    if ind == -1:
                        newscope.local('%s = _PyCmplNoType()' % _sanitizeParam(param))
                    else:
                        newscope.local('%s = %s' % (_sanitizeParam(param[:ind]), 
                                                    _sanitize(param[ind+1:])))

            decls = scp.subscopes + scp.locals
            decls.sort(key=lambda x: x[0])
            for decl in [d[1] for d in decls]:
                if isinstance(decl, Scope):
                    newscope.add(decl.Clone(0))
                else:
                    newscope.local(decl)
                
        self.currentscope = newscope
        return self.currentscope

    def parse(self, text, curline=0):
        """Parse the given text
        @param text: python code text to parse
        @keyword curline: current line of cursor for context

        """
        self.curline = curline
        buf = StringIO(text)
        self.gen = tokenize.generate_tokens(buf.readline)
        self.currentscope = self.scope

        try:
            freshscope = True
            while True:
                tokentype, token, indent = self.next()
                if tokentype == DEDENT or token == "pass":
                    self.scope = self.scope.pop(indent)
                elif token == 'def':
                    func = self._parsefunction(indent)
                    if func == None:
                        continue
                    freshscope = True
                    self.scope = self.scope.add(func)
                elif token == 'class':
                    cls = self._parseclass(indent)
                    if cls == None:
                        continue
                    freshscope = True
                    self.scope = self.scope.add(cls)
                    
                elif token == 'import':
                    imports = self._parseimportlist()
                    for mod, alias in imports:
                        loc = "import %s" % mod
                        if len(alias) > 0:
                            loc += " as %s" % alias
                        self.scope.local(loc)
                    freshscope = False
                elif token == 'from':
                    mod, token = self._parsedotname()
                    if not mod or token != "import":
                        continue
                    names = self._parseimportlist()
                    for name, alias in names:
                        loc = "from %s import %s" % (mod, name)
                        if len(alias) > 0:
                            loc += " as %s" % alias
                        self.scope.local(loc)
                    freshscope = False

                elif tokentype == STRING:
                    if freshscope:
                        self.scope.doc(token)
                elif tokentype == NAME:
                    name, token = self._parsedotname(token)
                    if token == '=':
                        stmt = self._parseassignment()
                        dbg("[pycomp] parseassignment: %s = %s" % (name, stmt))
                        if stmt != None:
                            # XXX Safety Check don't allow assignments to 
                            # item attributes unless the attribute is self
                            if u'.' not in name or name.startswith('self.'):
                                self.scope.local("%s = %s" % (name, stmt))
                    freshscope = False
        except StopIteration: #thrown on EOF
            pass
        except:
            dbg("[pycomp][err] Pyparser.parse: %s, %s" %
                (sys.exc_info()[0], sys.exc_info()[1]))
        return self._adjustvisibility()

#-----------------------------------------------------------------------------#
# Utility Functions
def _cleanstr(doc):
    """Clean up a docstring by removing quotes
    @param doc: docstring to clean up

    """
    return doc.replace('"', ' ').replace("'", ' ')

def _sanitize(cstr):
    """Sanitize a command string for namespace lookup
    @param cstr: command string to cleanup

    """
    val = ''
    level = 0
    for char in cstr:
        if char in ('(', '{', '['):
            level += 1
        elif char in (']', '}', ')'):
            level -= 1
        elif level == 0:
            val += char
    return val

def _sanitizeParam(param):
    """Cleanup a value string
    value = foo
    to ensure that 'value' is syntactically correct.

    """
    for i, c in enumerate(param):
        if c.isalpha() or c == '_':
            return param[i:]
    else:
        return param # something is wrong so let it bomb




###############################################################################
# Name: mssql.py                                                              #
# Purpose: Define Microsoft SQL syntax for highlighting and other features    #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: mssql.py                                                              
AUTHOR: Cody Precord                                                        
@summary: Lexer configuration module for Microsoft SQL.
@todo: too many to list                                                     

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _mssql.py 68355 2011-07-24 20:06:07Z CJP $"
__revision__ = "$Revision: 68355 $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc

# Local Imports
import syndata

#-----------------------------------------------------------------------------#

#---- Keyword Specifications ----#

# Data Types
MSSQL_DAT = (0, "")
# System Tables
MSSQL_SYS = (1, "")
# Global Variables
MSSQL_GLOB = (2, "")
# Functions
MSSQL_FUNC = (3, "")
# System Stored Procedures
MSSQL_SYSP = (4, "")
# Operators
MSSQL_OPS = (5, "")

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [ (stc.STC_MSSQL_DEFAULT, 'default_style'),
                 (stc.STC_MSSQL_COMMENT, 'comment_style'),
                 (stc.STC_MSSQL_COLUMN_NAME, 'keyword_style'),
                 (stc.STC_MSSQL_COLUMN_NAME_2, 'keyword_style'),
                 (stc.STC_MSSQL_DATATYPE, 'keyword2_style'),
                 (stc.STC_MSSQL_DEFAULT_PREF_DATATYPE, 'class_style'),
                 (stc.STC_MSSQL_FUNCTION, 'keyword3_style'),
                 (stc.STC_MSSQL_GLOBAL_VARIABLE, 'global_style'),
                 (stc.STC_MSSQL_IDENTIFIER, 'default_style'),
                 (stc.STC_MSSQL_LINE_COMMENT, 'comment_style'),
                 (stc.STC_MSSQL_NUMBER, 'number_style'),
                 (stc.STC_MSSQL_OPERATOR, 'operator_style'),
                 (stc.STC_MSSQL_STATEMENT, 'keyword_style'),
                 (stc.STC_MSSQL_STORED_PROCEDURE, 'scalar2_style'),
                 (stc.STC_MSSQL_STRING, 'string_style'),
                 (stc.STC_MSSQL_SYSTABLE, 'keyword4_style'),
                 (stc.STC_MSSQL_VARIABLE, 'scalar_style') ]

#---- Extra Properties ----#
FOLD = ("fold", "1")
FOLD_COMMENT = ("fold.comment", "1")
FOLD_COMPACT = ("fold.compact", "1")

#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for MS SQL""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_MSSQL)

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return SYNTAX_ITEMS

    def GetProperties(self):
        """Returns a list of Extra Properties to set """
        return [FOLD, FOLD_COMPACT]

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code """
        return [u'--']



###############################################################################
# Name: profiler.py                                                           #
# Purpose: Editra's user profile services                                     #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
This module provides the profile object and support functions for loading and
saving user preferences between sessions. The preferences are saved on disk as
a cPickle, because of this no objects that cannot be resolved in the namespace
of this module prior to starting the mainloop must not be put in the Profile as
it will cause errors on load. Ths means that only builtin python types should
be used and that a translation from that type to the required type should
happen during run time.

@summary: Editra's user profile management

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: profiler.py 71278 2012-04-25 15:47:19Z CJP $"
__revision__ = "$Revision: 71278 $"

#--------------------------------------------------------------------------#
# Imports
import os
import sys
import cPickle
import wx

# Editra Imports
from ed_glob import CONFIG, PROG_NAME, VERSION, PRINT_BLACK_WHITE, EOL_MODE_LF
import util
import dev_tool
import ed_msg

_ = wx.GetTranslation
#--------------------------------------------------------------------------#
# Globals
_DEFAULTS = {
           'ALPHA'      : 255,              # Transparency level (0-255)
           'AALIASING'  : False,            # Use Anti-Aliasing if availble
           'APPSPLASH'  : True,             # Show splash at startup
           'AUTOBACKUP' : False,            # Automatically backup files
           'AUTOBACKUP_PATH' : '',          # Backup path
           'AUTOBACKUP_SUFFIX' : '.edbkup', # Backup suffix (e.g. .edbkup, ~)
           'AUTO_COMP'  : True,             # Use Auto-comp if available
           'AUTO_COMP_EX' : False,          # Use extended autocompletion
           'AUTO_INDENT': True,             # Use Auto Indent
           'AUTO_TRIM_WS' : False,          # Trim whitespace on save
           'AUTO_RELOAD' : False,           # Automatically reload files?
           'BRACKETHL'  : True,             # Use bracket highlighting
           'BSUNINDENT' : True,             # Backspace Unindents
           'CARETWIDTH' : 1,                # Preferred cursor width
           'CHECKMOD'   : True,             # Auto check file for file mod
           'CHECKUPDATE': True,             # Check for updates on start
           'CODE_FOLD'  : True,             # Use code folding
           'CTRLBAR'    : dict(),           # ControlBar layouts
           'DEFAULT_LEX': 'Plain Text',     # Default lexer for new documents
           'DEFAULT'    : False,            # No longer used I believe
           'DEFAULT_VIEW' : 'Automatic',    # Default Perspective
           'EDGE'       : 80,               # Edge guide column
           'ENCODING'   : None,             # Prefered text encoding
           'EOL_MODE'   : EOL_MODE_LF,      # EOL mode 1 == LF, 2 == CRLF
           'FHIST'      : list(),           # List of history files
           'FHIST_LVL'  : 9,                # Filehistory length (9 is max)
           'FFILTER'    : 0,                # Last file filter used
           'GUIDES'     : False,            # Use Indentation guides
           'HLCARETLINE': False,            # Highlight Caret Line
           'ICONS'      : 'Tango',          # Icon Theme
           'ICON_SZ'    : (16, 16),         # Toolbar Icon Size
           'INDENTWIDTH': 4,                # Default indent width
           'ISBINARY'   : False,            # Is this instance a binary
           'KEY_PROFILE': None,             # Keybinding profile
           'LANG'       : 'Default',        # UI language
           'LASTCHECK'  : 0,                # Last time update check was done
           #'LEXERMENU'  : [lang_name,]     # Created on an as needed basis
           'MAXIMIZED'  : False,            # Was window maximized on exit
           'MODE'       : 'CODE',           # Overall editor mode
           'MYPROFILE'  : 'default.ppb',    # Path to profile file
           'OPEN_NW'    : False,            # Open files in new windows
           'PRINT_MODE' : PRINT_BLACK_WHITE,# Printer rendering mode
           'PROXY_SETTINGS' : dict(),       # Proxy Server Settings
           'REPORTER'   : True,             # Error Reporter is Active
           'SAVE_POS'   : True,             # Remember Carat positions
           'SAVE_SESSION' : False,          # Load previous session on startup
           'SEARCH_LOC' : list(),           # Recent Search Locations
           'SEARCH_FILTER' : '',            # Last used search filter
           'SESSION_KEY' : '',              # Ipc Session Server Key
           'SET_WPOS'   : True,             # Remember window position
           'SET_WSIZE'  : True,             # Remember mainwindow size on exit
           'SHOW_EDGE'  : True,             # Show Edge Guide
           'SHOW_EOL'   : False,            # Show EOL markers
           'SHOW_LN'    : True,             # Show Line Numbers
           'SHOW_WS'    : False,            # Show whitespace markers
           'SPELLCHECK' : dict(auto=False,
                               dict='en_US',
                               epath=None), # Spell checking preferences
           'STATBAR'    : True,             # Show Status Bar
           'SYNTAX'     : True,             # Use Syntax Highlighting
           'SYNTHEME'   : 'Default',        # Syntax Highlight color scheme
           'TABICONS'   : True,             # Show Tab Icons
           'TABWIDTH'   : 8,                # Tab width
           'THEME'      : 'DEFAULT',        # For future use
           'TOOLBAR'    : True,             # Show Toolbar
           'USETABS'    : False,            # Use tabs instead of spaces
           'USE_PROXY'  : False,            # Use Proxy server settings?
           'VIEWVERTSPACE' : False,         # Allow extra virtual space in buffer
           'VI_EMU'     : False,            # Use Vi emulation mode
           'VI_NORMAL_DEFAULT' : False,     # Use Normal mode by default
           'WARN_EOL'   : True,             # Warn about mixed eol characters
           'WRAP'       : False,            # Use Wordwrap
           'WSIZE'      : (700, 450)        # Mainwindow size
          #FONT1 created at runtime by ed_styles as primary font
          #FONT2 created at runtime by ed_styles as secondary font
          #FONT3 Standard Font by UI, added to profile when customized
}

#--------------------------------------------------------------------------#

class Profile(dict):
    """Class for managing profile data. All data is stored as builtin
    python objects (i.e. str, tuple, list, ect...) however on a request
    for data the object can be transformed in to a requested type where
    applicable. The profile saves itself to disk using the cPickle module
    to preserve data types and allow for easy loading.

    """
    _instance = None
    _created = False

    def __init__(self):
        """Initialize the profile"""
        if not self._created:
            dict.__init__(self)
        else:
            pass

    def __new__(cls, *args, **kargs):
        """Maintain only a single instance of this object
        @return: instance of this class

        """
        if cls._instance is None:
            cls._instance = dict.__new__(cls, *args, **kargs)
        return cls._instance

    #---- End Private Members ----#

    #---- Begin Public Members ----#
    def DeleteItem(self, item):
        """Removes an entry from the profile
        @param item: items name

        """
        if item in self:
            del self[item]
        else:
            pass

    def Get(self, index, fmt=None, default=None):
        """Gets the specified item from the data set

        @param index: index of item to get
        @keyword fmt: format the item should be in
        @keyword default: Default value to return if index is
                          not in profile.

        """
        if index in self:
            val = self.__getitem__(index)
        else:
            return default

        if fmt is None:
            return val
        else:
            return _ToObject(index, val, fmt)

    def Load(self, path):
        """Load the profiles data set with data from the given file
        @param path: path to file to load data from
        @note: The files data must have been written with a pickler

        """
        if os.path.exists(path):
            try:
                fhandle = open(path, 'rb')
                val = cPickle.load(fhandle)
                fhandle.close()
            except (IOError, SystemError, OSError,
                    cPickle.UnpicklingError, EOFError), msg:
                dev_tool.DEBUGP("[profile][err] %s" % str(msg))
            else:
                if isinstance(val, dict):
                    self.update(val)
                    self.Set('MYPROFILE', path)
                    dev_tool.DEBUGP("[profile][info] Loaded %s" % path)
        else:
            dev_tool.DEBUGP("[profile][err] %s does not exist" % path)
            dev_tool.DEBUGP("[profile][info] Loading defaults")
            self.LoadDefaults()
            self.Set('MYPROFILE', path)
            return False

        # Update profile to any keys that are missing
        for key in _DEFAULTS:
            if key not in self:
                self.Set(key, _DEFAULTS[key])
        return True

    def LoadDefaults(self):
        """Loads the default values into the profile
        @return: None

        """
        self.clear()
        self.update(_DEFAULTS)

    def Set(self, index, val, fmt=None):
        """Set the value of the given index
        @param index: Index to set
        @param val: Value to set
        @keyword fmt: Format to convert to string from

        """
        if fmt is None:
            self.__setitem__(index, val)
        else:
            self.__setitem__(index, _FromObject(val, fmt))

        # Notify all clients with the configuration change message
        ed_msg.PostMessage(ed_msg.EDMSG_PROFILE_CHANGE + (index,), val) 

    def Write(self, path):
        """Write the dataset of this profile as a pickle
        @param path: path to where to write the pickle
        @return: True on success / False on failure

        """
        try:
            # Only write if given an absolute path
            if not os.path.isabs(path):
                return False
            self.Set('MYPROFILE', path)
            fhandle = open(path, 'wb')
            cPickle.dump(self.copy(), fhandle, cPickle.HIGHEST_PROTOCOL)
            fhandle.close()
            UpdateProfileLoader()
        except (IOError, cPickle.PickleError), msg:
            dev_tool.DEBUGP(u"[profile][err] %s" % msg)
            return False
        else:
            return True

    def Update(self, update=None):
        """Update the profile using data from provided dictionary
        or the default set if none is given.
        @keyword update: dictionary of values to update from or None
        @postcondition: All profile values from the update set are set
                        in this profile. If update is None then the current
                        set is only updated to include values from the
                        DEFAULTS that are not currently present.

        """
        if update is None:
            for key, val in _DEFAULTS.iteritems():
                if key not in self:
                    self.Set(key, val)
        else:
            self.update(update)

    #---- End Public Members ----#

#-----------------------------------------------------------------------------#
# Singleton reference instance

TheProfile = Profile()

#-----------------------------------------------------------------------------#
# Profile convenience functions
Profile_Del  = TheProfile.DeleteItem
Profile_Get = TheProfile.Get
Profile_Set = TheProfile.Set

def _FromObject(val, fmt):
    """Convert the given value to a to a profile compatible value
    @param val: value to convert
    @param fmt: Format to convert to (string)

    """
    if fmt == u'font' and isinstance(val, wx.Font):
        return "%s,%s" % (val.GetFaceName(), val.GetPointSize())
    else:
        return val

def _ToObject(index, val, fmt):
    """Convert the given value to a different object
    @param index: fallback to retrieve item from defaults
    @param val: value to convert
    @param fmt: Format to convert to
    @todo: exception handling,

    """
    tmp = fmt.lower()
    if tmp == u'font':
        fnt = val.split(',')
        rval = wx.FFont(int(fnt[1]), wx.DEFAULT, face=fnt[0])
    elif tmp == u'bool':
        if isinstance(val, bool):
            rval = val
        else:
            rval = _DEFAULTS.get(index, False)
    elif tmp == u'size_tuple':
        if len(val) == 2 and \
           isinstance(val[0], int) and isinstance(val[1], int):
            rval = val
        else:
            rval = _DEFAULTS.get(index, wx.DefaultSize)
    elif tmp == u'str':
        rval = unicode(val)
    elif tmp == u'int':
        if isinstance(val, int):
            rval = val
        elif isinstance(val, basestring) and val.isdigit():
            rval = int(val)
        else:
            rval = _DEFAULTS.get(index)
    else:
        return val
    return rval

#---- Begin Function Definitions ----#

def CalcVersionValue(ver_str="0.0.0"):
    """Calculates a version value from the provided dot-formated string

    1) SPECIFICATION: Version value calculation AA.BBB.CCC
         - major values: < 1     (i.e 0.0.85 = 0.850)
         - minor values: 1 - 999 (i.e 0.1.85 = 1.850)
         - micro values: >= 1000 (i.e 1.1.85 = 1001.850)

    @keyword ver_str: Version string to calculate value of

    """
    ver_str = ''.join([char for char in ver_str
                       if char.isdigit() or char == '.'])
    ver_lvl = ver_str.split(u".")
    if len(ver_lvl) < 3:
        return 0

    major = int(ver_lvl[0]) * 1000
    minor = int(ver_lvl[1])
    if len(ver_lvl[2]) <= 2:
        ver_lvl[2] += u'0'
    micro = float(ver_lvl[2]) / 1000
    return float(major) + float(minor) + micro

def GetLoader():
    """Finds the loader to use
    @return: path to profile loader
    @note: path may not exist, only returns the path to where the loader
           should be.

    """
    cbase = util.GetUserConfigBase()
    loader = os.path.join(cbase, u"profiles", u".loader2")
    return loader

def GetProfileStr():
    """Reads the profile string from the loader and returns it.
    The profile string must be the first line in the loader file.
    @return: path of profile used in last session

    """
    reader = util.GetFileReader(GetLoader())
    if reader == -1:
        # So return the default
        return CONFIG['PROFILE_DIR'] + u"default.ppb"

    profile = reader.readline()
    profile = profile.strip()
    reader.close()
    if not os.path.isabs(profile):
        profile = CONFIG['PROFILE_DIR'] + profile
    return profile

def ProfileIsCurrent():
    """Checks if profile is compatible with current editor version
    and returns a bool stating if it is or not.
    @return: whether profile on disk was written with current program version

    """
    if CalcVersionValue(ProfileVersionStr()) >= CalcVersionValue(VERSION):
        return True
    else:
        return False

def ProfileVersionStr():
    """Checks the Loader for the profile version string and
    returns the version string. If there is an error or the
    string is not found it returns a zero version string.
    @return: the version string value from the profile loader file

    """
    loader = GetLoader()
    reader = util.GetFileReader(loader, sys.getfilesystemencoding())
    if reader == -1:
        return "0.0.0"

    ret_val = "0.0.0"
    count = 0
    while True:
        count += 1
        value = reader.readline()
        value = value.split()
        if len(value) > 0:
            if value[0] == u'VERSION':
                ret_val = value[1]
                break
        # Give up after 20 lines if version string not found
        if count > 20:
            break
    reader.close()

    return ret_val

def UpdateProfileLoader():
    """Updates Loader File
    @precondition: MYPROFILE has been set
    @postcondition: on disk profile loader is updated
    @return: 0 if no error, non zero for error condition

    """
    writer = util.GetFileWriter(GetLoader())
    if writer == -1:
        return 1

    prof_name = Profile_Get('MYPROFILE')
    if not prof_name or not os.path.isabs(prof_name):
        prof_name = CONFIG['PROFILE_DIR'] + 'default.ppb'

    if not os.path.exists(prof_name):
        prof_name = os.path.join(CONFIG['CONFIG_DIR'],
                                 os.path.basename(prof_name))
        Profile_Set('MYPROFILE', prof_name)

    # Use just the relative profile name for local(portable) config paths
    prof_name = os.path.basename(prof_name)

    writer.write(prof_name)
    writer.write(u"\nVERSION\t" + VERSION)
    writer.close()
    return 0



###############################################################################
# Name: backupmgr.py                                                          #
# Purpose: File Backup Manager                                                #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2009 Cody Precord <staff@editra.org>                         #
# Licence: wxWindows Licence                                                  #
###############################################################################

"""
Editra Business Model Library: FileBackupMgr

Helper class for managing and creating backups of files.

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__cvsid__ = "$Id: backupmgr.py 67646 2011-04-29 03:07:20Z CJP $"
__revision__ = "$Revision: 67646 $"

__all__ = [ 'FileBackupMgr', ]

#-----------------------------------------------------------------------------#
# Imports
import os
import shutil

# Local Imports
import fileutil
import fchecker

#-----------------------------------------------------------------------------#

class FileBackupMgr(object):
    """File backup creator and manager"""
    def __init__(self, header=None, template=u"%s~"):
        """Create a BackupManager
        @keyword header: header to id backups with (Text files only!!)
        @keyword template: template string for naming backup file with

        """
        super(FileBackupMgr, self).__init__()

        # Attributes
        self.checker = fchecker.FileTypeChecker()
        self.header = header       # Backup id header
        self.template = template   # Filename template
        self.bkupdir = u""

    def _CheckHeader(self, fname):
        """Check if the backup file has a header that matches the
        header used to identify backup files.
        @param fname: name of file to check
        @return: bool (True if header is ok, False otherwise)

        """
        isok = False
        handle = None
        try:
            handle = open(fname, 'r')
            line = handle.readline()
            handle.close()
            isok = line.startswith(self.header)
        except Exception, msg:
            isok = False
            if handle:
                handle.close()
        return isok

    def GetBackupFilename(self, fname):
        """Get the unique name for the files backup copy
        @param fname: string (file path)
        @return: string

        """
        if self.bkupdir:
            tmp = fileutil.GetFileName(fname)
            fname = os.path.join(self.bkupdir, tmp)

        rname = self.template % fname
        if self.header is not None and \
           not self.checker.IsBinary(fname) and \
           os.path.exists(rname):
            # Make sure that the template backup name does not match
            # an existing file that is not a backup file.
            while os.path.exists(rname):
                if not self._CheckHeader(rname):
                    rname = self.template % rname
                else:
                    break
                
        return rname

    def GetBackupWriter(self, fileobj):
        """Create a backup filewriter method to backup a files contents
        with.
        @param fileobj: object implementing fileimpl.FileObjectImpl interface
        @return: callable(text) to create backup with

        """
        nfile = fileobj.Clone()
        fname = self.GetBackupFilename(nfile.GetPath())
        nfile.SetPath(fname)
        # Write the header if it is enabled
        if self.header and not self.checker.IsBinary(fname):
            nfile.Write(self.header + os.linesep)
        return nfile.Write

    def HasBackup(self, fname):
        """Check if a given file has a backup file available or not
        @param fname: string (file path)

        """
        backup = self.GetBackupFilename(fname)
        return os.path.exists(backup)

    def IsBackupNewer(self, fname):
        """Is the backup of this file newer than the saved version
        of the file?
        @param fname: string (file path)
        @return: bool

        """
        backup = self.GetBackupFilename(fname)
        if os.path.exists(fname) and os.path.exists(backup):
            mod1 = fileutil.GetFileModTime(backup)
            mod2 = fileutil.GetFileModTime(fname)
            return mod1 > mod2
        else:
            return False

    def MakeBackupCopy(self, fname):
        """Create a backup copy of the given filename
        @param fname: string (file path)
        @return: bool (True == Success)

        """
        backup = self.GetBackupFilename(fname)
        try:
            if os.path.exists(backup):
                os.remove(backup)

            shutil.copy2(fname, backup)
            if self.header:
                handle = open(backup, 'r')
                txt = handle.read()
                handle.close()
                handle = open(backup, 'w')
                handle.write(self.header + os.linesep)
                handle.write(txt)
                handle.close()
        except:
            return False
        else:
            return True

    def SetBackupDirectory(self, path):
        """Set the backup directory to use for all backups created by this
        manager instance. Setting the path to an empty string will set the
        default behavior to write the backup to the same directory as the
        where the file that is being backedup is located.

        """
        self.bkupdir = path 

    def SetBackupFileTemplate(self, tstr):
        """Set the filename template for generating the backupfile name
        @param tstr: template string i.e) %s~

        """
        assert tstr.count("%s") == 1, "Format statment must only have one arg"
        self.template = tstr

    def SetHeader(self, header):
        """Set the header string for identifying a file as a backup
        @param header: string (single line only)

        """
        assert '\n' not in header, "Header must only be a single line"
        self.header = header



###############################################################################
# Name: testThreadpool.py                                                     #
# Purpose: Unit tests for the Icon Theme provider                             #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2011 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""Unittest cases for testing the icon theme provider interface"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id:  $"
__revision__ = "$Revision:  $"

#-----------------------------------------------------------------------------#
# Imports
import wx
import os
import time
import threading
import unittest

# Module to test
import ebmlib

#-----------------------------------------------------------------------------#
# Test Class

def doWork(seconds, id_, callback):
    """Worker job for threadpool test"""
    time.sleep(seconds)
    callback(id_) # notify done

class ThreadpoolTest(unittest.TestCase):
    """Tests the ThreadPool class"""
    def setUp(self):
        self._threads = 3
        self.pool = ebmlib.ThreadPool(self._threads)
        self._completed = list()
        self._lock = threading.Lock()

    def tearDown(self):
        self.pool.Shutdown()
        self._completed = list()

    def notifyDone(self, id_):
        with self._lock:
            self._completed.append(id_)

    #---- Test Cases ----#

    def testThreadCount(self):
        """Check that thread pool spawned off requested number of threads"""
        self.assertEquals(self.pool.ThreadCount, self._threads)

    def testQueueJob(self):
        jobs = [ (1.5, 1), (2, 2), (0.5, 3), (1, 4) ]
        for job in jobs:
            self.pool.QueueJob(doWork, job[0], job[1], self.notifyDone)
        maxwait = sum([j[0] for j in jobs])
        time.sleep(maxwait) # wait for the jobs to finish
        self.assertTrue(len(jobs) == len(self._completed), repr(self._completed))



###############################################################################
# Name: ed_keyh.py                                                            #
# Purpose: Editra's Vi Emulation Key Handler                                  #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
KeyHandler interface for implementing extended key action handling in Editra's
main text editting buffer.

@summary: Custom keyhandler interface

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: ed_keyh.py 70747 2012-02-29 01:33:35Z CJP $"
__revision__ = "$Revision: 70747 $"

#-------------------------------------------------------------------------#
# Imports
import re
import wx
import wx.stc

# Editra Libraries
import ed_event
import ed_glob
import ed_basestc
import ed_stc
import string
import ed_vim

#-------------------------------------------------------------------------#
# Use this base class to derive any new keyhandlers from. The keyhandler is
# called upon by the active buffer when ever a key press event happens. The
# handler then has the responsibility of deciding what to do with the key.
#
class KeyHandler(object):
    """KeyHandler base class"""
    def __init__(self, stc):
        super(KeyHandler, self).__init__()

        # Attributes
        self.stc = stc
        self._blockmode = False

    STC = property(lambda self: self.stc)
    BlockMode = property(lambda self: self._blockmode,
                         lambda self,m:setattr(self,'_blockmode', m))

    def ClearMode(self):
        """Clear any key input modes to normal input mode"""
        evt = ed_event.StatusEvent(ed_event.edEVT_STATUS, self.stc.Id,
                                   '', ed_glob.SB_BUFF)
        wx.PostEvent(self.stc.TopLevelParent, evt)

    def GetHandlerName(self):
        """Get the name of this handler
        @return: string

        """
        return u'NULL'

    def PreProcessKey(self, key_code, ctrldown=False,
                      cmddown=False, shiftdown=False, altdown=False):
        """Pre process any keys before they get to the char handler
        @param key_code: Raw keycode
        @keyword ctrldown: Is the control key down
        @keyword cmddown: Is the Command key down (Mac osx)
        @keyword shiftdown: Is the Shift key down
        @keyword altdown: Is the Alt key down
        @return: bool

        """
        return False

    def ProcessKey(self, key_code, ctrldown=False,
                   cmddown=False, shiftdown=False, altdown=False):
        """Process the key and return True if it was processed and
        false if it was not. The key is recieved at EVT_CHAR.
        @param key_code: Raw keycode
        @keyword ctrldown: Is the control key down
        @keyword cmddown: Is the Command key down (Mac osx)
        @keyword shiftdown: Is the Shift key down
        @keyword altdown: Is the Alt key down
        @return: bool

        """
        return False

#-------------------------------------------------------------------------#

class ViKeyHandler(KeyHandler):
    """Defines a key handler for Vi emulation
    @summary: Handles key presses according to Vi emulation.

    """

    # Vi Mode declarations
    NORMAL, \
    INSERT, \
    VISUAL, \
        = range(3)

    def __init__(self, stc, use_normal_default=False):
        super(ViKeyHandler, self).__init__(stc)

        # Attributes
        self.mode = 0
        self.last = u''
        self.last_find = u''
        self.commander = ed_vim.EditraCommander(self)
        self.buffer = u''

        # Insert mode by default
        if use_normal_default:
            self.NormalMode()
        else:
            self.InsertMode()

    def ClearMode(self):
        """Clear the mode back to default input mode"""
        # TODO:CJP when newer scintilla is available in 2.9 use
        #          blockcaret methods.
        self.STC.SetLineCaret()
        self.BlockMode = False
        self.last = self.cmdcache = u''
        super(ViKeyHandler, self).ClearMode()

    def GetHandlerName(self):
        """Get the name of this handler"""
        return u'VI'

    def _SetMode(self, newmode, msg):
        """Set the keyhandlers mode
        @param newmode: New mode name to change to

        """
        self.buffer = u'' # Clear buffer from last mode
        self.mode = newmode
        # Update status bar
        evt = ed_event.StatusEvent(ed_event.edEVT_STATUS, self.stc.GetId(),
                                   msg, ed_glob.SB_BUFF)
        wx.PostEvent(self.stc.GetTopLevelParent(), evt)

    def InsertMode(self):
        """Change to insert mode"""
        self.stc.SetLineCaret()
        self.stc.SetOvertype(False)
        self.BlockMode = False
        self._SetMode(ViKeyHandler.INSERT, u"INSERT")

    def ReplaceMode(self):
        """Change to replace mode
        This really just insert mode with overtype set to true

        """
        self.stc.SetLineCaret()
        self.stc.SetOvertype(True)
        self._SetMode(ViKeyHandler.INSERT, u"REPLACE")

    def NormalMode(self):
        """Change to normal (command) mode"""
        if self.IsInsertMode():
            self.commander.SetLastInsertedText(self.buffer)

        self.stc.SetOvertype(False)
        self.stc.SetBlockCaret()
        self.BlockMode = True
        self.commander.Deselect()
        self.commander.InsertRepetition()
        self._SetMode(ViKeyHandler.NORMAL, u'NORMAL')

    def VisualMode(self):
        """Change to visual (selection) mode"""
        self.stc.SetBlockCaret()
        self.BlockMode = True
        self.stc.SetOvertype(False)
        self._SetMode(ViKeyHandler.VISUAL, u'VISUAL')
        self.commander.StartSelection()

    def IsInsertMode(self):
        """Test if we are in insert mode"""
        return self.mode == ViKeyHandler.INSERT

    def IsNormalMode(self):
        """Test if we are in normal mode"""
        return self.mode == ViKeyHandler.NORMAL

    def IsVisualMode(self):
        """Test if we are in visual mode"""
        return self.mode == ViKeyHandler.VISUAL

    def PreProcessKey(self, key_code, ctrldown=False,
                      cmddown=False, shiftdown=False, altdown=False):
        """Pre process any keys before they get to the char handler
        @param key_code: Raw keycode
        @keyword ctrldown: Is the control key down
        @keyword cmddown: Is the Command key down (Mac osx)
        @keyword shiftdown: Is the Shift key down
        @keyword altdown: Is the Alt key down
        @return: bool

        """
        if not shiftdown and key_code == wx.WXK_ESCAPE:
            # If Vi emulation is active go into Normal mode and
            # pass the key event to the char handler by not processing
            # the key.
            self.NormalMode()
            return False
        elif (ctrldown or cmddown) and key_code == ord('['):
            self.NormalMode()
            return True
        elif key_code in (wx.WXK_RETURN, wx.WXK_BACK,
                          wx.WXK_RIGHT, wx.WXK_LEFT) and \
             not self.IsInsertMode():
            # swallow enter key in normal and visual modes
            # HACK: we have to do it form here because ProcessKey
            #       is only called on Char events, not Key events,
            #       and the Enter key does not generate a Char event.
            self.ProcessKey(key_code)
            return True
        else:
            return False

    def ProcessKey(self, key_code, ctrldown=False,
                   cmddown=False, shiftdown=False, altdown=False):
        """Processes keys and decided whether to interpret them as vim commands
        or normal insert text

        @param key_code: Raw key code
        @keyword cmddown: Command/Ctrl key is down
        @keyword shiftdown: Shift Key is down
        @keyword altdown : Alt key is down

        """
        if ctrldown or cmddown or altdown:
            return False

        # Mode may change after processing the key, so we need to remember
        # whether this was a command or not
        f_cmd = self.IsNormalMode() or self.IsVisualMode()

        self._ProcessKey(key_code)

        # Update status bar
        if self.IsNormalMode():
            if self.stc.GetTopLevelParent():
                evt = ed_event.StatusEvent(ed_event.edEVT_STATUS,
                                           self.stc.GetId(),
                                           u"NORMAL %s" % self.buffer,
                                           ed_glob.SB_BUFF)
                wx.PostEvent(self.stc.GetTopLevelParent(), evt)

        if f_cmd:
            return True
        else:
            # If we're in insert mode we must return False
            # so the text gets inserted into the editor
            return False

    def _ProcessKey(self, key_code):
        """The real processing of keys"""
        char = unichr(key_code)
        if self.IsNormalMode() or self.IsVisualMode():
            self.buffer += char
            if ed_vim.Parse(self.buffer, self.commander):
                # command was handled (or invalid) so clear buffer
                self.buffer = u''

            if self.IsVisualMode():
                self.commander.ExtendSelection()

        elif self.IsInsertMode():
            self.buffer += char

    def InsertText(self, pos, text):
        """Insert text and store it in the buffer if we're in insert mode
        i.e. as if it was typed in

        """
        self.stc.InsertText(pos, text)
        if self.IsInsertMode():
            self.buffer += text


###############################################################################
# Name: ed_print.py                                                           #
# Purpose: Editra's printer class                                             #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
Printer class for creating and managing printouts from a StyledTextCtrl.

Classes:
  - L{EdPrinter}: Class for managing printing and providing print dialogs
  - L{EdPrintout}: Scales and renders the given document to a printer.

@summary: Printer Classes for printing text from an STC

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__cvsid__ = "$Id: ed_print.py 67499 2011-04-15 20:33:40Z CJP $"
__revision__ = "$Revision: 67499 $"

#--------------------------------------------------------------------------#
# Imports
import wx
import wx.stc

# Editra Imports
import ed_glob
import util
import extern.stcprint as stcprint

_ = wx.GetTranslation
#--------------------------------------------------------------------------#

# Globals
COLOURMODES = { ed_glob.PRINT_BLACK_WHITE : wx.stc.STC_PRINT_BLACKONWHITE,
                ed_glob.PRINT_COLOR_WHITE : wx.stc.STC_PRINT_COLOURONWHITE,
                ed_glob.PRINT_COLOR_DEF   : wx.stc.STC_PRINT_COLOURONWHITEDEFAULTBG,
                ed_glob.PRINT_INVERT      : wx.stc.STC_PRINT_INVERTLIGHT,
                ed_glob.PRINT_NORMAL      : wx.stc.STC_PRINT_NORMAL }

#--------------------------------------------------------------------------#
class EdPrinter(object):
    """Printer Class for the editor
    @note: current font size is fixed at 12 point for printing

    """
    def __init__(self, parent, mode=ed_glob.PRINT_NORMAL):
        """Initializes the Printer
        @param parent: parent window
        @keyword mode: printer mode

        """
        super(EdPrinter, self).__init__()

        # Attributes
        self.stc = None
        self.title = wx.EmptyString
        self.parent = parent
        self.print_mode = mode
        self.print_data = wx.PrintData()
        self.margins = (wx.Point(15,15), wx.Point(15,15))

    def CreatePrintout(self):
        """Creates a printout of the current stc window
        @return: a printout object

        """
        colour = COLOURMODES[self.print_mode]
        dlg_data = wx.PageSetupDialogData(self.print_data)
        dlg_data.SetPrintData(self.print_data)
        dlg_data.SetMarginTopLeft(self.margins[0])
        dlg_data.SetMarginBottomRight(self.margins[1])
        fname = self.stc.GetFileName()
        printout = stcprint.STCPrintout(self.stc, page_setup_data=dlg_data, 
                                        print_mode=colour, title=self.title,
                                        job_title=fname)
        return printout

    def PageSetup(self):
        """Opens a print setup dialog and save print settings.
        @return: None

        """
        dlg_data = wx.PageSetupDialogData(self.print_data)
        dlg_data.SetPrintData(self.print_data)
    
        dlg_data.SetDefaultMinMargins(True)
        dlg_data.SetMarginTopLeft(self.margins[0])
        dlg_data.SetMarginBottomRight(self.margins[1])

        print_dlg = wx.PageSetupDialog(self.parent, dlg_data)
        if print_dlg.ShowModal() == wx.ID_OK:
            self.print_data = wx.PrintData(dlg_data.GetPrintData())
            self.print_data.SetPaperId(dlg_data.GetPaperId())
            self.margins = (dlg_data.GetMarginTopLeft(),
                            dlg_data.GetMarginBottomRight())
        print_dlg.Destroy()

    def Preview(self):
        """Preview the Print
        @return: None

        """
        printout = self.CreatePrintout()
        printout2 = self.CreatePrintout()
        preview = wx.PrintPreview(printout, printout2, self.print_data)
        preview.SetZoom(150)
        if preview.IsOk():
            pre_frame = wx.PreviewFrame(preview, self.parent,
                                             _("Print Preview"))
            dsize = wx.GetDisplaySize()
            pre_frame.SetInitialSize((self.stc.GetSize()[0],
                                          dsize.GetHeight() - 100))
            pre_frame.Initialize()
            pre_frame.Show()
        else:
            wx.MessageBox(_("Failed to create print preview"),
                          _("Print Error"),
                          style=wx.ICON_ERROR|wx.OK)

    def Print(self):
        """Prints the document
        @postcondition: the current document is printed

        """
        pdd = wx.PrintDialogData(self.print_data)
        printer = wx.Printer(pdd)
        printout = self.CreatePrintout()
        result = printer.Print(self.parent, printout)
        if result:
            dlg_data = printer.GetPrintDialogData()
            self.print_data = wx.PrintData(dlg_data.GetPrintData())
        elif printer.GetLastError() == wx.PRINTER_ERROR:
            wx.MessageBox(_("There was an error when printing.\n"
                            "Check that your printer is properly connected."),
                          _("Printer Error"),
                          style=wx.ICON_ERROR|wx.OK)
        printout.Destroy()

    def SetColourMode(self, mode):
        """Sets the color mode that the text is to be rendered with
        @param mode: mode to set the printer to use
        @return: whether mode was set or not
        @rtype: boolean

        """
        if mode in COLOURMODES:
            self.print_mode = mode
            ret = True
        else:
            ret = False
        return ret

    def SetStc(self, stc):
        """Set the stc we are printing for
        @param stc: instance of wx.stc.StyledTextCtrl
        @note: MUST be called prior to any other print operations

        """
        self.stc = stc



###############################################################################
# Name: miscutil.py                                                           #
# Purpose: Various helper functions.                                          #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2009 Cody Precord <staff@editra.org>                         #
# Licence: wxWindows Licence                                                  #
###############################################################################

"""
Editra Business Model Library: MiscUtil

Various helper functions

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__cvsid__ = "$Id: miscutil.py 67329 2011-03-28 23:40:48Z CJP $"
__revision__ = "$Revision: 67329 $"

__all__ = [ 'MinMax', 'Singleton']

#-----------------------------------------------------------------------------#
# Imports

#-----------------------------------------------------------------------------#

class Singleton(type):
    """Singleton metaclass for creating singleton classes
    @note: class being applied to must have a SetupWindow method

    """
    def __init__(cls, name, bases, dict):
        super(Singleton, cls).__init__(name, bases, dict)
        cls.instance = None

    def __call__(cls, *args, **kw):
        if not cls.instance:
            # Not created or has been Destroyed
            obj = super(Singleton, cls).__call__(*args, **kw)
            cls.instance = obj

        return cls.instance

#-----------------------------------------------------------------------------#

def MinMax(arg1, arg2):
    """Return an ordered tuple of the minimum and maximum value
    of the two args.
    @return: tuple

    """
    return min(arg1, arg2), max(arg1, arg2)
